# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""Provide a context to run a device without a database."""

# Imports
import os
import time
import struct
import socket
import tempfile
import traceback
import collections
import psutil
from functools import partial

# Concurrency imports
import threading
import multiprocessing
import queue

# CLI imports
from ast import literal_eval
from importlib import import_module
from argparse import ArgumentParser, ArgumentTypeError

# Local imports
from tango.server import run
from tango.utils import (
    _clear_test_context_tango_host_fqtrl,
    is_non_str_seq,
    _set_test_context_tango_host_fqtrl,
)
from tango.green import switch_existing_global_executors_to_thread
from tango import Database, DevFailed, DeviceProxy, EnsureOmniThread, Util

__all__ = (
    "MultiDeviceTestContext",
    "DeviceTestContext",
    "run_device_test_context",
    "get_server_port_via_pid",
)

# Helpers

_DEFAULT_THREAD_TIMEOUT = 5.0
_DEFAULT_PROCESS_TIMEOUT = 7.0

IOR = collections.namedtuple(
    "IOR",
    "first dtype_length dtype nb_profile tag "
    "length major minor wtf host_length host port body",
)

NO_DB_FRAGMENT = "dbase=no"


def ascii_to_bytes(s):
    convert = lambda x: bytes((int(x, 16),))
    return b"".join(convert(s[i : i + 2]) for i in range(0, len(s), 2))


def parse_ior(encoded_ior):
    assert encoded_ior[:4] == "IOR:"
    ior = ascii_to_bytes(encoded_ior[4:])
    dtype_length = struct.unpack_from("II", ior)[-1]
    form = f"II{dtype_length:d}sIIIBBHI"
    host_length = struct.unpack_from(form, ior)[-1]
    form = f"II{dtype_length:d}sIIIBBHI{host_length:d}sH0I"
    values = struct.unpack_from(form, ior)
    values += (ior[struct.calcsize(form) :],)
    strip = lambda x: x[:-1] if isinstance(x, bytes) else x
    return IOR(*map(strip, values))


def get_server_host_port():
    util = Util.instance()
    ds = util.get_dserver_device()
    encoded_ior = util.get_dserver_ior(ds)
    ior = parse_ior(encoded_ior)
    return ior.host.decode(), ior.port


def get_server_port_via_pid(pid, host, retries=400, delay=0.03):
    """Return the TCP port that a device server process is listening on (GIOP).

    This checks TCP sockets open on the process with the given PID, and attempts
    to find the one that accepts GIOP traffic.  A connection will be made to each
    listening socket, and data may be sent to them.

    General Inter-ORB Protocol (GIOP) is the message protocol which object
    request brokers (ORBs) communicate in CORBA.  This port is the one that is
    used when connecting a DeviceProxy.  These are not the port(s) used for ZMQ
    event traffic.

    :param pid: operating system process identifier
    :type pid: int
    :param host: hostname/IP that device server is listening on.  E.g., 127.0.0.1,
                 IP address of a non-loopback network interface, etc.  Note that starting a device
                 server on "localhost" may fail if OmniORB creates an IPv6-only socket.
    :type host: str
    :param retries: number of times to retry attempts, optional
    :type retries: int
    :param delay: time to wait (seconds) between retries, optional
    :type delay: float

    :returns: TCP port number
    :rtype: int

    :raises RuntimeError: If the GIOP port couldn't be identified

    .. versionadded:: 9.4.0
    .. versionadded:: 9.5.0
        *retries* parameter.
        *delay* parameter.
    """

    count = 0
    port = None
    last_err = None
    while port is None and count < retries:
        ports = _get_listening_tcp_ports(pid)
        try:
            port = _get_giop_port(host, ports)
        except Exception as err:
            last_err = err
            time.sleep(delay)
        count += 1

    if port is None:
        raise RuntimeError(
            f"Failed to get GIOP TCP port within {count * delay:.1f} sec"
        ) from last_err

    return port


def _get_listening_tcp_ports(pid):
    p = psutil.Process(pid)
    if hasattr(p, "net_connections"):
        conns = p.net_connections(kind="tcp")
    else:
        conns = p.connections(kind="tcp")  # deprecated in psutil v6.0.0
    return list(set([c.laddr.port for c in conns if c.status == "LISTEN"]))


def _get_giop_port(host, ports):
    protocols = _try_get_protocols_on_ports(host, ports)
    for port, protocol in protocols.items():
        if protocol == "GIOP":
            return port
    raise RuntimeError(
        f"None of ports {ports} appear to have GIOP protocol. "
        f"Guessed protocols: {protocols}."
    )


def _try_get_protocols_on_ports(host, ports):
    """Return a dict with port to protocol mapping.

    This attempts to establish a TCP socket connection to the host
    for each port, and then determine the protocol it supports.

    ZMQ client sockets receive an unsolicited version check message on connection.
    CORBA GIOP client sockets don't receive an unsolicited message, so we send
    a requested to disconnect, and expect an empty message back.
    """
    zmq_response = (
        b"\xff\x00\x00\x00\x00\x00\x00\x00\x01\x7f"  # signature + version check
    )
    giop_send = b"GIOP\x01\x02\x01\x05\x00\x00\x00\x00"  # request disconnect
    giop_response = b""  # graceful disconnect
    max_bytes_expected = len(zmq_response)

    protocols = dict.fromkeys(ports, "Unknown")
    for port in ports:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            sock.settimeout(0.001)
            server_address = (host, port)
            sock.connect(server_address)

            try:
                data = sock.recv(max_bytes_expected)
                if data == zmq_response:
                    protocols[port] = "ZMQ"
                    continue
            except OSError:
                pass

            try:
                sock.sendall(giop_send)
                data = sock.recv(max_bytes_expected)
                if data == giop_response:
                    protocols[port] = "GIOP"
                    continue
            except OSError:
                pass
        except OSError:
            pass
        finally:
            sock.close()
    return protocols


def literal_dict(arg):
    return dict(literal_eval(arg))


def device(path):
    """Get the device class from a given module."""
    module_name, device_name = path.rsplit(".", 1)
    try:
        module = import_module(module_name)
    except Exception:
        raise ArgumentTypeError(
            f"Error importing {module_name}.{device_name}:\n"
            f"{traceback.format_exc()}"
        )
    return getattr(module, device_name)


def get_host_ip():
    """Get the primary external host IP.

    This is used to be useful because an explicit IP was required to get
    tango events to work properly.  Now remains for backwards compatibility.
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    # Connecting to a UDP address doesn't send packets
    # Note:  For some reason Python3 on macOS does not accept 0 as a port but
    # returns with an errno 49 instead. Therefore, just use port 80 which just
    # works as well.
    s.connect(("8.8.8.8", 80))
    # Get ip address
    ip = s.getsockname()[0]
    s.close()
    return ip


def _device_class_from_field(field):
    """
    Helper function that extracts and return a device class from a
    'class' field of a devices_info dictionary

    :param field: the field from which to extract the device class

    :return: the device class extracted from the field
    """
    device_cls_class = None
    if is_non_str_seq(field):
        (device_cls_class, device_class) = (field[0], field[1])
        if isinstance(device_cls_class, str):
            device_cls_class = device(device_cls_class)
    else:
        device_class = field
    if isinstance(device_class, str):
        device_class = device(device_class)
    return (device_cls_class, device_class)


class MultiDeviceTestContext:
    """Context to run device(s) without a database.

    The difference with respect to
    :class:`~tango.test_context.DeviceTestContext` is that it allows
    to export multiple devices (even of different Tango classes).

    Example usage::

        from tango import DeviceProxy
        from tango.server import Device, attribute
        from tango.test_context import MultiDeviceTestContext


        class Device1(Device):
            @attribute(dtype=int)
            def attr1(self):
                return 1


        class Device2(Device):
            @attribute(dtype=int)
            def attr2(self):
                dev1 = DeviceProxy("test/device/1")
                return dev1.attr1 * 2


        devices_info = (
            {
                "class": Device1,
                "devices": [
                    {"name": "test/device/1"},
                ],
            },
            {
                "class": Device2,
                "devices": [
                    {
                        "name": "test/device/2",
                    },
                ],
            },
        )


        def test_devices():
            with MultiDeviceTestContext(devices_info, process=True) as context:
                proxy1 = context.get_device("test/device/1")
                proxy2 = context.get_device("test/device/2")
                assert proxy1.attr1 == 1
                assert proxy2.attr2 == 2

    :param devices_info:
      a sequence of dicts with information about
      devices to be exported. Each dict consists of the following keys:

        * "class" which value is either of:

          * a :class:`~tango.server.Device` or the name of some such class
          * a sequence of two elements, the first element being a
            :class:`~tango.DeviceClass` or the name of some such class,
            the second element being a :class:`~tango.DeviceImpl` or the
            name of some such class

        * "devices" which value is a sequence of dicts with the following keys:

          * "name" (str)
          * "properties" (dict)
          * "memorized" (dict)
          * "root_atts" (dict)"

    :type devices_info:
      sequence<dict>
    :param server_name:
      Name to use for the device server.
      Optional.  Default is the first device's class name.
    :type server_name:
      :py:obj:`str`
    :param instance_name:
      Name to use for the device server instance.
      Optional.  Default is lower-case version of the server name.
    :type instance_name:
      :py:obj:`str`
    :param db:
      Path to a pre-populated text file to use for the
      database.
      Optional.  Default is to create a new temporary file and populate it
      based on the devices and properties supplied in `devices_info`.
    :type db:
      :py:obj:`str`
    :param host:
      Hostname to use for device server's ORB endpoint.
      Optional.  Default is the loopback IP address, 127.0.0.1.
    :type host:
      :py:obj:`str`
    :param port:
      Port number to use for the device server's ORB endpoint.
      Optional.  Default is chosen by omniORB.
    :type port:
      :py:obj:`int`
    :param debug:
      Debug level for the device server logging.
      0=OFF, 1=FATAL, 2=ERROR, 3=WARN, 4=INFO, 5=DEBUG.
      Optional. Default is warn.
    :type debug:
      :py:obj:`int`
    :param process:
      True if the device server should be launched in a new process, otherwise
      use a new thread.  Note:  if the context will be used mutiple times, it
      may seg fault if the thread mode is chosen.
      See the :ref:`issues <testing-approaches-issues>` and
      :ref:`process kwarg <testing-approaches-process-kwarg>` discussion in the docs.
      Optional.  Default is thread.
    :type process:
      :py:obj:`bool`
    :param daemon:
      True if the new thread/process must be created in daemon mode.
      Optional.  Default is not daemon.
    :type daemon:
      :py:obj:`bool`
    :param timeout:
      How long to wait (seconds) for the device server to start up, and also
      how long to wait on joining the thread/process when stopping.
      Optional.  Default differs for thread and process modes.
    :type timeout:
      :py:obj:`float`
    :param green_mode:
      Green mode to use for the device server.
      Optional.  Default uses the Device specification (via green_mode class attribute),
      or if that isn't specified the global green mode.
    :type green_mode:
      :obj:`~tango.GreenMode`

    .. versionadded:: 9.3.2

    .. versionadded:: 9.3.3
        Added support for `memorized` key to "devices" field in `devices_info`.
        Added support for literal names for "class" field in `devices_info`.

    .. versionadded:: 9.3.3
        added *green_mode* parameter.

    .. versionchanged:: 9.5.0
        By default, devices launched by a test context can be accessed using
        short names with :class:`~tango.AttributeProxy`, :class:`~tango.DeviceProxy`,
        and :class:`~tango.Group`.
        This can be disabled by setting the `enable_test_context_tango_host_override`
        class/instance attribute to `False` before starting the test context.
        Added support for `root_atts` key to "devices" field in `devices_info`.
    """

    command = "{0} {1} -ORBendPoint giop:tcp:{2}:{3} -file={4}"
    enable_test_context_tango_host_override = True

    thread_timeout = _DEFAULT_THREAD_TIMEOUT
    process_timeout = _DEFAULT_PROCESS_TIMEOUT

    def __init__(
        self,
        devices_info,
        server_name=None,
        instance_name=None,
        db=None,
        host=None,
        port=0,
        debug=3,
        process=False,
        daemon=False,
        timeout=None,
        green_mode=None,
    ):
        if not server_name:
            _, first_device = _device_class_from_field(devices_info[0]["class"])
            server_name = first_device.__name__
        if not instance_name:
            instance_name = server_name.lower()
        if db is None:
            handle, db = tempfile.mkstemp()
            self.handle = handle
        else:
            self.handle = None
        if host is None:
            host = "127.0.0.1"  # note: localhost does not currently work
        if timeout is None:
            timeout = self.process_timeout if process else self.thread_timeout
        # Attributes
        self.db = db
        self.host = host
        self.port = port
        self.timeout = timeout
        self.server_name = "/".join(("dserver", server_name, instance_name))
        if process:
            self._startup_exception_queue = multiprocessing.Queue()
            self._discovered_port_queue = multiprocessing.JoinableQueue()
        else:
            self._startup_exception_queue = queue.Queue()
            self._discovered_port_queue = queue.Queue()
        self._startup_exception = None
        self._devices = {}
        self._saved_environ = {}
        self._process = process

        # Command args
        string = self.command.format(server_name, instance_name, host, port, db)
        string += f" -v{debug}" if debug else ""
        cmd_args = string.split()

        class_list = []
        device_list = []
        tangoclass_list = []
        for device_info in devices_info:
            device_cls, device = _device_class_from_field(device_info["class"])
            tangoclass = device.__name__
            if tangoclass in tangoclass_list:
                self.delete_db()
                raise ValueError(
                    "multiple entries in devices_info pointing "
                    "to the same Tango class"
                )
            tangoclass_list.append(tangoclass)
            # File
            self.append_db_file(
                server_name, instance_name, tangoclass, device_info["devices"]
            )
            if device_cls:
                class_list.append((device_cls, device, tangoclass))
            else:
                device_list.append(device)

        # Target and arguments
        if class_list and device_list:
            self.delete_db()
            raise ValueError(
                "mixing HLAPI and classical API in devices_info is not supported"
            )
        if class_list:
            runserver = partial(run, class_list, cmd_args, green_mode=green_mode)
        elif len(device_list) == 1 and hasattr(device_list[0], "run_server"):
            runserver = partial(device.run_server, cmd_args, green_mode=green_mode)
        elif device_list:
            runserver = partial(run, device_list, cmd_args, green_mode=green_mode)
        else:
            raise ValueError("Wrong format of devices_info")

        cls = multiprocessing.Process if process else threading.Thread
        self.thread = cls(target=self.target, args=(runserver, process))
        self.thread.daemon = daemon

    def target(self, runserver, process=False):
        with EnsureOmniThread():
            threading.current_thread().name += " TestContext DS launcher"
            try:
                runserver(
                    pre_init_callback=self.pre_init,
                    post_init_callback=self.post_init,
                    raises=True,
                )
            except Exception as exc:
                self._startup_exception_queue.put(exc)
                self._discovered_port_queue.put(-1)  # don't block queue reader
            finally:
                # Put something in the queue just in case
                exc = RuntimeError("The server failed to report anything")
                self._startup_exception_queue.put(exc)
                # Make sure the process has enough time to send the items
                # because it might segfault while cleaning up the tango resources
                if process:
                    time.sleep(0.1)

    def pre_init(self):
        try:
            if not self.port:
                util = Util.instance()
                pid = util.get_pid()
                self.port = get_server_port_via_pid(pid, self.host)
            if self._process:
                # set TANGO_HOST override for device server (DS) launcher child process
                self._override_test_context_tango_host()
            # report it to the test context process (DS launcher's parent)
            self._discovered_port_queue.put(self.port)
            # and wait till it has been processed
            self._discovered_port_queue.join()
        except Exception as exc:
            self._startup_exception_queue.put(exc)
            self._discovered_port_queue.put(-1)  # don't block queue reader

    def _wait_until_port_is_known(self):
        try:
            self.port = self._discovered_port_queue.get(timeout=self.timeout)
        except queue.Empty:
            raise RuntimeError(
                "GIOP TCP port of TextContext device server not available during startup"
            )

        # set the TANGO_HOST override for the process that created the test context
        self._override_test_context_tango_host()
        # notify the server process, so it can continue
        self._discovered_port_queue.task_done()

    def _override_test_context_tango_host(self):
        if self.enable_test_context_tango_host_override:
            _set_test_context_tango_host_fqtrl(
                f"tango://{self.host}:{self.port}#{NO_DB_FRAGMENT}"
            )

    def post_init(self):
        # now we can connect to the device - success!
        self._startup_exception_queue.put(None)

    def append_db_file(self, server, instance, tangoclass, device_prop_info):
        """Generate a database file corresponding to the given arguments."""
        device_names = [info["name"] for info in device_prop_info]
        # Open the file
        with open(self.db, "a") as f:
            f.write("/".join((server, instance, "DEVICE", tangoclass)))
            f.write(": ")
            f.write(", ".join(device_names))
            f.write("\n")
        # Create database
        db = Database(self.db)
        # Write properties
        for info in device_prop_info:
            device_name = info["name"]
            properties = dict(info.get("properties", {}))
            # Patch the property dict to avoid a PyTango bug
            for key, value in properties.items():
                if is_non_str_seq(value):
                    properties[key] = [v if v != "" else " " for v in value]
                else:
                    properties[key] = value if value != "" else " "
            db.put_device_property(device_name, properties)

            root_atts = info.get("root_atts", {})
            properties_to_save = {
                attribute_name: {"__root_att": root_att}
                for (attribute_name, root_att) in root_atts.items()
            }

            memorized = info.get("memorized", {})
            for attribute_name, memorized_value in memorized.items():
                if attribute_name in properties_to_save:
                    properties_to_save[attribute_name]["__value"] = memorized_value
                else:
                    properties_to_save[attribute_name] = {"__value": memorized_value}

            db.put_device_attribute_property(device_name, properties_to_save)
        try:
            validated_db = Database(self.db)
        except DevFailed:
            with open(self.db, "r") as f:
                content = f.read()
            raise RuntimeError(
                f"Invalid FileDatabase file was created.\n"
                f"Check device properties (empty list or str?): "
                f"{device_prop_info}.\n"
                f"Problematic file has content:\n{content}"
            )
        return validated_db

    def delete_db(self):
        """delete temporary database file only if it was created by this class"""
        if self.handle is not None:
            os.close(self.handle)
            os.unlink(self.db)

    def get_server_access(self) -> str:
        """Return the full server name."""
        return f"tango://{self.host}:{self.port}/{self.server_name}#{NO_DB_FRAGMENT}"

    def get_device_access(self, device_name) -> str:
        """Return the full device name."""
        return f"tango://{self.host}:{self.port}/{device_name}#{NO_DB_FRAGMENT}"

    def get_device(self, device_name) -> DeviceProxy:
        """Return the device proxy corresponding to the given device name.

        Maintains previously accessed device proxies in a cache to not recreate
        then on every access.
        """
        if device_name not in self._devices:
            device = DeviceProxy(self.get_device_access(device_name))
            self._devices[device_name] = device
        return self._devices[device_name]

    def start(self):
        """Run the server.

        This method is automatically called when the context handler is entered.

        :raises RuntimeError: If device server does not start
        """
        self._set_up_environment_variables()
        if not self._process:
            self.thread.daemon = True
        self.thread.start()
        self.connect()
        return self

    def connect(self):
        try:
            self._wait_until_port_is_known()
            self._wait_until_startup_status_is_known()
        except RuntimeError:
            if self.thread.is_alive():
                raise RuntimeError(
                    "The server appears to be stuck at initialization. "
                    "Check stdout/stderr for more information."
                )
            elif hasattr(self.thread, "exitcode"):
                raise RuntimeError(
                    f"The server process stopped with exitcode {self.thread.exitcode}. "
                    f"Check stdout/stderr for more information."
                )
            else:
                raise RuntimeError(
                    "The server stopped without reporting. "
                    "Check stdout/stderr for more information."
                )

        if self._startup_exception:
            raise self._startup_exception

        # Get server proxy
        self.server = DeviceProxy(self.get_server_access())
        self.server.ping()

        # If server has green_mode, we should switch global executor to the main thread
        if not self._process:
            switch_existing_global_executors_to_thread()

    def _wait_until_startup_status_is_known(self):
        try:
            self._startup_exception = self._startup_exception_queue.get(
                timeout=self.timeout
            )
        except queue.Empty:
            raise RuntimeError(
                f"Timed-out waiting for device server startup ({self.timeout:1.3f} sec)"
            )

    def stop(self):
        """Kill the server.

        This method is automatically called when the context handler is exited.

        :raises RuntimeError: If device server does not stop cleanly

        .. versionchanged:: 10.1.0
            Can raise :class:`RuntimeError` on shutdown.
        """
        try:
            if self.server:
                self.server.command_inout("Kill")
            self.join(self.timeout)
            self._ensure_device_server_exited()
        finally:
            _clear_test_context_tango_host_fqtrl()
            self.delete_db()
            self._restore_environment_variables()

    def _ensure_device_server_exited(self):
        if not self.thread.is_alive():
            return
        if self._process:
            self.thread.kill()
            t_start = time.monotonic()
            while self.thread.is_alive() and time.monotonic() - t_start < self.timeout:
                time.sleep(0.1)
            raise RuntimeError(
                f"Device server failed to exit cleanly (stuck in shutdown?). "
                f"Tried to kill subprocess. "
                f"Still alive: {self.thread.is_alive()}."
            )
        else:
            raise RuntimeError(
                "Device server failed to exit cleanly (stuck in shutdown?)"
            )

    def join(self, timeout=None):
        self.thread.join(timeout)

    def _set_up_environment_variables(self):
        self._saved_environ = dict(os.environ)
        if self._process:
            # This variable is used by omniORB to tell it how often
            # to check for idle connections (which can be closed).
            # See: https://omniorb.net/omni43/omniORB/omniORB006.html#sec103
            # It is also used for various timeouts when shutting down.
            # We are interested in the shutdown case.
            # The default is 5 seconds, so we reduce it to get faster
            # process shutdown (especially noticeable on Windows)
            # We can't use 0 seconds because omniORB forces it to 5 seconds.
            # See: src/lib/omniORB/orbcore/giopServer.cc giopServer::deactivate()
            # See also: https://sourceforge.net/p/tango-cs/bugs/819/
            os.environ["ORBscanGranularity"] = "1"

    def _restore_environment_variables(self):
        modified_environ = dict(os.environ)
        saved_environ = self._saved_environ
        for key, value in modified_environ.items():
            if key in saved_environ:
                saved_value = saved_environ[key]
                if value != saved_value:
                    os.environ[key] = saved_value
            else:
                os.environ.pop(key)  # unset environment var

    def __enter__(self) -> "MultiDeviceTestContext":
        """Enter method for context handler support.

        :return:
          Instance of this test context.  Use `get_device` to get proxy
          access to any of the devices started by this context.
        :rtype:
          :class:`~tango.test_context.MultiDeviceTestContext`

        :raises RuntimeError: If device server does not start
        """
        if not self.thread.is_alive():
            self.start()
        return self

    def __exit__(self, exc_type, exception, trace):
        """Exit method for context handler support.

        :raises RuntimeError: If device server does not stop cleanly

        .. versionchanged:: 10.1.0
            Can raise :class:`RuntimeError` on shutdown.
        """
        self.stop()
        return False


# Single device test context
class DeviceTestContext(MultiDeviceTestContext):
    """Context to run a single device without a database.

    The difference with respect to
    :class:`~tango.test_context.MultiDeviceTestContext` is that it only
    allows to export a single device.

    Example usage::

        from time import sleep

        from tango.server import Device, attribute, command
        from tango.test_context import DeviceTestContext

        class PowerSupply(Device):

            @attribute(dtype=float)
            def voltage(self):
                return 1.23

            @command
            def calibrate(self):
                sleep(0.1)

        def test_calibrate():
            '''Test device calibration and voltage reading.'''
            with DeviceTestContext(PowerSupply, process=True) as proxy:
                proxy.calibrate()
                assert proxy.voltage == 1.23

    :param device:
      Device class to be run.
    :type device:
      :class:`~tango.server.Device` or :class:`~tango.DeviceImpl`
    :param device_cls:
      The device class can be provided if using the low-level API.
      Optional.  Not required for high-level API devices, of type
      :class:`~tango.server.Device`.
    :type device_cls:
      :class:`~tango.DeviceClass`

     The rest of the parameters are described in
     :class:`~tango.test_context.MultiDeviceTestContext`.

    .. versionadded:: 9.2.1

    .. versionadded:: 9.3.3
        added *memorized* parameter.

    .. versionadded:: 9.3.6
        added *green_mode* parameter.
    """

    def __init__(
        self,
        device,
        device_cls=None,
        server_name=None,
        instance_name=None,
        device_name=None,
        properties=None,
        db=None,
        host=None,
        port=0,
        debug=3,
        process=False,
        daemon=False,
        timeout=None,
        memorized=None,
        root_atts=None,
        green_mode=None,
    ):
        # Argument
        if not server_name:
            server_name = device.__name__
        if not instance_name:
            instance_name = server_name.lower()
        if not device_name:
            device_name = "test/nodb/" + server_name.lower()
        if properties is None:
            properties = {}
        if memorized is None:
            memorized = {}
        if root_atts is None:
            root_atts = {}
        if device_cls:
            cls = (device_cls, device)
        else:
            cls = device
        devices_info = (
            {
                "class": cls,
                "devices": (
                    {
                        "name": device_name,
                        "properties": properties,
                        "memorized": memorized,
                        "root_atts": root_atts,
                    },
                ),
            },
        )
        super().__init__(
            devices_info,
            server_name=server_name,
            instance_name=instance_name,
            db=db,
            host=host,
            port=port,
            debug=debug,
            process=process,
            daemon=daemon,
            timeout=timeout,
            green_mode=green_mode,
        )

        self.device_name = device_name
        self.device = self.server = None

    def get_device_access(self, device_name=None) -> str:
        """Return the full device name."""
        if device_name is None:
            device_name = self.device_name
        return super().get_device_access(device_name)

    def connect(self):
        super().connect()
        # Get device proxy
        self.device = self.get_device(self.device_name)
        self.device.ping()

    def __enter__(self) -> DeviceProxy:
        """Enter method for context handler support.

        :return:
          A device proxy to the device started by this context.
        :rtype:
          :class:`~tango.DeviceProxy`
        """
        if not self.thread.is_alive():
            self.start()
        return self.device


# Command line interface


def parse_command_line_args(args=None):
    """Parse arguments given in command line."""
    desc = "Run a given device on a given port."
    parser = ArgumentParser(description=desc)
    # Add arguments
    msg = "The device to run as a python path."
    parser.add_argument("device", metavar="DEVICE", type=device, help=msg)
    msg = "The hostname to use."
    parser.add_argument(
        "--host", metavar="HOST", type=str, help=msg, default=get_host_ip()
    )
    msg = "The port to use."
    parser.add_argument("--port", metavar="PORT", type=int, help=msg, default=8888)
    msg = "The debug level."
    parser.add_argument("--debug", metavar="DEBUG", type=int, help=msg, default=3)
    msg = "The properties to set as python dict."
    parser.add_argument(
        "--prop", metavar="PROP", type=literal_dict, help=msg, default="{}"
    )
    msg = "Whether to disable short-name lookup for devices launched by test context"
    parser.add_argument(
        "--disable-short-name-lookup",
        dest="disable_short_name_lookup",
        action="store_true",
        help=msg,
    )
    # Parse arguments
    namespace = parser.parse_args(args)
    return (
        namespace.device,
        namespace.host,
        namespace.port,
        namespace.prop,
        namespace.debug,
        namespace.disable_short_name_lookup,
    )


def run_device_test_context(args=None):
    (
        device_,
        host,
        port,
        properties,
        debug,
        disable_short_name_lookup,
    ) = parse_command_line_args(args)
    context = DeviceTestContext(
        device_, properties=properties, host=host, port=port, debug=debug
    )
    context.enable_test_context_tango_host_override = not disable_short_name_lookup
    context.start()
    msg = f"{device_.__name__} started on port {context.port} with properties {properties}"
    print(msg)
    print(f"Device access: {context.get_device_access()}")
    print(f"Server access: {context.get_server_access()}")
    context.join()
    print("Done")


# Main execution

if __name__ == "__main__":
    run_device_test_context()



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
from tango import (
    Util,
    SerialModel,
    AutoTangoAllowThreads,
    EnsureOmniThread,
    AutoTangoMonitor,
    DevFailed,
)
from tango.server import Device, attribute
from tango.test_utils import DeviceTestContext

from threading import Thread, Event
from queue import Queue

import pytest

from time import sleep

pytestmark = pytest.mark.extra_src_test


def do_nothing():
    sleep(1e-3)


class SimpleDS(Device):

    def init_device(self):
        u = Util.instance()
        mode = SerialModel.BY_CLASS

        u.set_serial_model(mode)
        assert u.get_serial_model() == mode

        with AutoTangoAllowThreads(self):
            with AutoTangoAllowThreads(self):
                do_nothing()

    @attribute
    def attr(self) -> int:
        return 1234


def test_dont_crash_with_nested_tango_allow_threads():
    with DeviceTestContext(SimpleDS) as proxy:
        assert proxy.attr == 1234


class MonSameThread(Device):
    serial_model = None
    unlock = None
    running = None

    def init_device(self):
        u = Util.instance()

        u.set_serial_model(MonSameThread.serial_model)
        assert u.get_serial_model() == MonSameThread.serial_model

        self._thread = Thread(target=self.thread_func)
        MonSameThread.running = True
        self._thread.start()

    def delete_device(self):
        MonSameThread.running = False
        self._thread.join()

    def thread_func(self):
        with EnsureOmniThread():
            if MonDiffThread.serial_model == SerialModel.BY_CLASS:
                obj = self.get_device_class()
            else:
                obj = self

        with AutoTangoMonitor(obj):
            if MonSameThread.unlock:
                # unlock it again
                with AutoTangoAllowThreads(self):
                    while MonSameThread.running:
                        do_nothing()
            else:
                while MonSameThread.running:
                    do_nothing()

    @attribute
    def attr(self) -> int:
        return 1234


@pytest.mark.parametrize("unlock", [True, False])
def test_monitor_force_unlock_from_same_thread(
    unlock, server_green_mode, server_serial_model
):
    MonSameThread.serial_model = server_serial_model
    MonSameThread.unlock = unlock

    if unlock:
        if server_serial_model == SerialModel.BY_CLASS:
            pytest.xfail(
                "Not implemented, see https://gitlab.com/tango-controls/pytango/-/issues/650"
            )
        elif server_serial_model == SerialModel.BY_PROCESS:
            pytest.xfail(
                "Not implemented, see https://gitlab.com/tango-controls/pytango/-/issues/650"
            )

    with DeviceTestContext(MonSameThread) as proxy:

        if server_serial_model == SerialModel.NO_SYNC:
            assert proxy.attr == 1234
        elif unlock:
            assert proxy.attr == 1234
        else:
            with pytest.raises(
                DevFailed, match="not able to acquire serialization monitor"
            ):
                assert proxy.attr == 1234

        # required especially by SerialModel.BY_PROCESS so that the DS can be killed from the test context
        MonSameThread.running = False


class MonDiffThread(Device):
    serial_model = None
    unlock = None
    running = None

    def init_device(self):
        u = Util.instance()

        u.set_serial_model(MonDiffThread.serial_model)
        assert u.get_serial_model() == MonDiffThread.serial_model

        MonDiffThread.running = True

        self._lock_thread = Thread(target=self.lock_thread_func)
        self._lock_thread.start()

        self._unlock_thread = Thread(target=self.unlock_thread_func)
        self._unlock_thread.start()

    def delete_device(self):
        MonDiffThread.running = False
        self._lock_thread.join()
        self._unlock_thread.join()

    def lock_thread_func(self):
        with EnsureOmniThread():
            if MonDiffThread.serial_model == SerialModel.BY_CLASS:
                obj = self.get_device_class()
            else:
                obj = self

            with AutoTangoMonitor(obj):
                while MonDiffThread.running:
                    do_nothing()

    def unlock_thread_func(self):
        with EnsureOmniThread():
            if MonDiffThread.unlock:
                # try to unlock it, but as this is from a different thread, nothing happens
                with AutoTangoAllowThreads(self):
                    while MonDiffThread.running:
                        do_nothing()
            else:
                while MonDiffThread.running:
                    do_nothing()

    @attribute
    def attr(self) -> int:
        return 1234


@pytest.mark.parametrize("unlock", [True, False])
def test_monitor_force_unlock_from_different_thread(
    unlock, server_green_mode, server_serial_model
):
    MonDiffThread.serial_model = server_serial_model
    MonDiffThread.unlock = unlock

    with DeviceTestContext(MonDiffThread) as proxy:

        if server_serial_model == SerialModel.NO_SYNC:
            assert proxy.attr == 1234
        else:
            with pytest.raises(
                DevFailed, match="not able to acquire serialization monitor"
            ):
                assert proxy.attr == 1234

        # required especially by SerialModel.BY_PROCESS so that the DS can be killed from the test context
        MonDiffThread.running = False


class DontDeadlock(Device):
    def init_device(self) -> None:
        self.queue = Queue[Event | None]()
        self.thread = Thread(target=self._run)
        self.thread.start()

    def delete_device(self) -> None:
        self.queue.put(None)
        self.thread.join()

    def _run(self) -> None:
        with EnsureOmniThread():
            while True:
                ev = self.queue.get()
                if ev is None:
                    break

                with AutoTangoMonitor(self):
                    ev.set()

    @attribute
    def attr(self) -> int:
        ev = Event()
        self.queue.put(ev)
        # Allow the thread to run and start waiting for the monitor
        sleep(0.01)
        with AutoTangoAllowThreads(self):
            ev.wait()

        return 1234


# This test will probably only fail if the system is under load
def test_dont_deadlock_when_releasing_monitor():
    with DeviceTestContext(DontDeadlock, process=True) as proxy:
        for _ in range(100):
            assert proxy.attr == 1234



#!/usr/bin/env python
# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
# -*- coding:utf-8 -*-


##############################################################################
## license :
##============================================================================
##
## File :        DataBase.py
##
## Project :     TANGO
##
## $Author :      controls$
##
## $Revision :    $
##
## $Date :        $
##
## $HeadUrl :     $
##============================================================================
##            This file is generated by POGO
##    (Program Obviously used to Generate tango Object)
##
##        (c) - Software Engineering Group - ESRF
##############################################################################

"""This class manage the TANGO database."""

__all__ = ("DataBase", "DataBaseClass", "main")

__docformat__ = "restructuredtext"

import PyTango
import sys

# Add additional import
# ----- PROTECTED REGION ID(DataBase.additionnal_import) ENABLED START -----#
try:
    from __future__ import print_function
except ImportError:
    pass
# ----- PROTECTED REGION END -----#	//	DataBase.additionnal_import

##############################################################################
## Device States Description
##
## No states for this device
##############################################################################


class DataBase(PyTango.Device_4Impl):
    # --------- Add you global variables here --------------------------
    # ----- PROTECTED REGION ID(DataBase.global_variables) ENABLED START -----#

    # ----- PROTECTED REGION END -----#	//	DataBase.global_variables
    # ------------------------------------------------------------------
    #    Device constructor
    # ------------------------------------------------------------------
    def __init__(self, cl, name):
        PyTango.Device_4Impl.__init__(self, cl, name)
        self.debug_stream("In " + self.get_name() + ".__init__()")
        DataBase.init_device(self)

    # ------------------------------------------------------------------
    #    Device destructor
    # ------------------------------------------------------------------
    def delete_device(self):
        self.debug_stream("In " + self.get_name() + ".delete_device()")
        # ----- PROTECTED REGION ID(DataBase.delete_device) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.delete_device

    # ------------------------------------------------------------------
    #    Device initialization
    # ------------------------------------------------------------------
    def init_device(self):
        self.debug_stream("In " + self.get_name() + ".init_device()")
        self.get_device_properties(self.get_device_class())
        self.attr_StoredProcedureRelease_read = ""
        self.attr_Timing_average_read = [0.0]
        self.attr_Timing_minimum_read = [0.0]
        self.attr_Timing_maximum_read = [0.0]
        self.attr_Timing_calls_read = [0.0]
        self.attr_Timing_index_read = [""]
        self.attr_Timing_info_read = [""]
        # ----- PROTECTED REGION ID(DataBase.init_device) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.init_device

    # ------------------------------------------------------------------
    #    Always excuted hook method
    # ------------------------------------------------------------------
    def always_executed_hook(self):
        self.debug_stream("In " + self.get_name() + ".always_excuted_hook()")
        # ----- PROTECTED REGION ID(DataBase.always_executed_hook) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.always_executed_hook

    # ==================================================================
    #
    #    DataBase read/write attribute methods
    #
    # ==================================================================

    # ------------------------------------------------------------------
    #    Read StoredProcedureRelease attribute
    # ------------------------------------------------------------------
    def read_StoredProcedureRelease(self, attr):
        self.debug_stream("In " + self.get_name() + ".read_StoredProcedureRelease()")
        # ----- PROTECTED REGION ID(DataBase.StoredProcedureRelease_read) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.StoredProcedureRelease_read
        attr.set_value(self.attr_StoredProcedureRelease_read)

    # ------------------------------------------------------------------
    #    Read Timing_average attribute
    # ------------------------------------------------------------------
    def read_Timing_average(self, attr):
        self.debug_stream("In " + self.get_name() + ".read_Timing_average()")
        # ----- PROTECTED REGION ID(DataBase.Timing_average_read) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.Timing_average_read
        attr.set_value(self.attr_Timing_average_read)

    # ------------------------------------------------------------------
    #    Read Timing_minimum attribute
    # ------------------------------------------------------------------
    def read_Timing_minimum(self, attr):
        self.debug_stream("In " + self.get_name() + ".read_Timing_minimum()")
        # ----- PROTECTED REGION ID(DataBase.Timing_minimum_read) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.Timing_minimum_read
        attr.set_value(self.attr_Timing_minimum_read)

    # ------------------------------------------------------------------
    #    Read Timing_maximum attribute
    # ------------------------------------------------------------------
    def read_Timing_maximum(self, attr):
        self.debug_stream("In " + self.get_name() + ".read_Timing_maximum()")
        # ----- PROTECTED REGION ID(DataBase.Timing_maximum_read) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.Timing_maximum_read
        attr.set_value(self.attr_Timing_maximum_read)

    # ------------------------------------------------------------------
    #    Read Timing_calls attribute
    # ------------------------------------------------------------------
    def read_Timing_calls(self, attr):
        self.debug_stream("In " + self.get_name() + ".read_Timing_calls()")
        # ----- PROTECTED REGION ID(DataBase.Timing_calls_read) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.Timing_calls_read
        attr.set_value(self.attr_Timing_calls_read)

    # ------------------------------------------------------------------
    #    Read Timing_index attribute
    # ------------------------------------------------------------------
    def read_Timing_index(self, attr):
        self.debug_stream("In " + self.get_name() + ".read_Timing_index()")
        # ----- PROTECTED REGION ID(DataBase.Timing_index_read) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.Timing_index_read
        attr.set_value(self.attr_Timing_index_read)

    # ------------------------------------------------------------------
    #    Read Timing_info attribute
    # ------------------------------------------------------------------
    def read_Timing_info(self, attr):
        self.debug_stream("In " + self.get_name() + ".read_Timing_info()")
        # ----- PROTECTED REGION ID(DataBase.Timing_info_read) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.Timing_info_read
        attr.set_value(self.attr_Timing_info_read)

    # ------------------------------------------------------------------
    #    Read Attribute Hardware
    # ------------------------------------------------------------------
    def read_attr_hardware(self, data):
        self.debug_stream("In " + self.get_name() + ".read_attr_hardware()")
        # ----- PROTECTED REGION ID(DataBase.read_attr_hardware) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.read_attr_hardware

    # ==================================================================
    #
    #    DataBase command methods
    #
    # ==================================================================

    # ------------------------------------------------------------------
    #    DbAddDevice command:
    # ------------------------------------------------------------------
    def DbAddDevice(self, argin):
        """Add a Tango class device to a specific device server

        :param argin: Str[0] = Full device server process name
        Str[1] = Device name
        Str[2] = Tango class name
        :type: PyTango.DevVarStringArray
        :return:
        :rtype: PyTango.DevVoid"""
        self.debug_stream("In " + self.get_name() + ".DbAddDevice()")
        # ----- PROTECTED REGION ID(DataBase.DbAddDevice) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbAddDevice

    # ------------------------------------------------------------------
    #    DbAddServer command:
    # ------------------------------------------------------------------
    def DbAddServer(self, argin):
        """Create a device server process entry in database

        :param argin: Str[0] = Full device server name
        Str[1] = Device(s) name
        Str[2] = Tango class name
        Str[n] = Device name
        Str[n + 1] = Tango class name
        :type: PyTango.DevVarStringArray
        :return:
        :rtype: PyTango.DevVoid"""
        self.debug_stream("In " + self.get_name() + ".DbAddServer()")
        # ----- PROTECTED REGION ID(DataBase.DbAddServer) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbAddServer

    # ------------------------------------------------------------------
    #    DbDeleteAttributeAlias command:
    # ------------------------------------------------------------------
    def DbDeleteAttributeAlias(self, argin):
        """Delete an attribute alias.

        :param argin: Attriibute alias name.
        :type: PyTango.DevString
        :return:
        :rtype: PyTango.DevVoid"""
        self.debug_stream("In " + self.get_name() + ".DbDeleteAttributeAlias()")
        # ----- PROTECTED REGION ID(DataBase.DbDeleteAttributeAlias) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbDeleteAttributeAlias

    # ------------------------------------------------------------------
    #    DbDeleteClassAttribute command:
    # ------------------------------------------------------------------
    def DbDeleteClassAttribute(self, argin):
        """delete a class attribute and all its properties from database

        :param argin: Str[0] = Tango class name
        Str[1] = Attribute name
        :type: PyTango.DevVarStringArray
        :return:
        :rtype: PyTango.DevVoid"""
        self.debug_stream("In " + self.get_name() + ".DbDeleteClassAttribute()")
        # ----- PROTECTED REGION ID(DataBase.DbDeleteClassAttribute) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbDeleteClassAttribute

    # ------------------------------------------------------------------
    #    DbDeleteClassAttributeProperty command:
    # ------------------------------------------------------------------
    def DbDeleteClassAttributeProperty(self, argin):
        """delete class attribute properties from database

        :param argin: Str[0] = Tango class name
        Str[1] = Attribute name
        Str[2] = Property name
        Str[n] = Property name
        :type: PyTango.DevVarStringArray
        :return:
        :rtype: PyTango.DevVoid"""
        self.debug_stream("In " + self.get_name() + ".DbDeleteClassAttributeProperty()")
        # ----- PROTECTED REGION ID(DataBase.DbDeleteClassAttributeProperty) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbDeleteClassAttributeProperty

    # ------------------------------------------------------------------
    #    DbDeleteClassProperty command:
    # ------------------------------------------------------------------
    def DbDeleteClassProperty(self, argin):
        """Delete class properties from database

        :param argin: Str[0] = Tango class name
        Str[1] = Property name
        Str[n] = Property name
        :type: PyTango.DevVarStringArray
        :return:
        :rtype: PyTango.DevVoid"""
        self.debug_stream("In " + self.get_name() + ".DbDeleteClassProperty()")
        # ----- PROTECTED REGION ID(DataBase.DbDeleteClassProperty) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbDeleteClassProperty

    # ------------------------------------------------------------------
    #    DbDeleteDevice command:
    # ------------------------------------------------------------------
    def DbDeleteDevice(self, argin):
        """Delete a devcie from database

        :param argin: device name
        :type: PyTango.DevString
        :return:
        :rtype: PyTango.DevVoid"""
        self.debug_stream("In " + self.get_name() + ".DbDeleteDevice()")
        # ----- PROTECTED REGION ID(DataBase.DbDeleteDevice) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbDeleteDevice

    # ------------------------------------------------------------------
    #    DbDeleteDeviceAlias command:
    # ------------------------------------------------------------------
    def DbDeleteDeviceAlias(self, argin):
        """Delete a device alias.

        :param argin: device alias name
        :type: PyTango.DevString
        :return:
        :rtype: PyTango.DevVoid"""
        self.debug_stream("In " + self.get_name() + ".DbDeleteDeviceAlias()")
        # ----- PROTECTED REGION ID(DataBase.DbDeleteDeviceAlias) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbDeleteDeviceAlias

    # ------------------------------------------------------------------
    #    DbDeleteDeviceAttribute command:
    # ------------------------------------------------------------------
    def DbDeleteDeviceAttribute(self, argin):
        """Delete  device attribute properties from database

        :param argin: Str[0] = Device name
        Str[1] = Attribute name
        :type: PyTango.DevVarStringArray
        :return:
        :rtype: PyTango.DevVoid"""
        self.debug_stream("In " + self.get_name() + ".DbDeleteDeviceAttribute()")
        # ----- PROTECTED REGION ID(DataBase.DbDeleteDeviceAttribute) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbDeleteDeviceAttribute

    # ------------------------------------------------------------------
    #    DbDeleteDeviceAttributeProperty command:
    # ------------------------------------------------------------------
    def DbDeleteDeviceAttributeProperty(self, argin):
        """delete a device attribute property from the database

        :param argin: Str[0] = Device name
        Str[1] = Attribute name
        Str[2] = Property name
        Str[n] = Property name
        :type: PyTango.DevVarStringArray
        :return:
        :rtype: PyTango.DevVoid"""
        self.debug_stream(
            "In " + self.get_name() + ".DbDeleteDeviceAttributeProperty()"
        )
        # ----- PROTECTED REGION ID(DataBase.DbDeleteDeviceAttributeProperty) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbDeleteDeviceAttributeProperty

    # ------------------------------------------------------------------
    #    DbDeleteDeviceProperty command:
    # ------------------------------------------------------------------
    def DbDeleteDeviceProperty(self, argin):
        """Delete device property(ies)

        :param argin: Str[0] = Device name
        Str[1] = Property name
        Str[n] = Property name
        :type: PyTango.DevVarStringArray
        :return:
        :rtype: PyTango.DevVoid"""
        self.debug_stream("In " + self.get_name() + ".DbDeleteDeviceProperty()")
        # ----- PROTECTED REGION ID(DataBase.DbDeleteDeviceProperty) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbDeleteDeviceProperty

    # ------------------------------------------------------------------
    #    DbDeleteProperty command:
    # ------------------------------------------------------------------
    def DbDeleteProperty(self, argin):
        """Delete free property from database

        :param argin: Str[0]  = Object name
        Str[1] = Property name
        Str[n] = Property name
        :type: PyTango.DevVarStringArray
        :return:
        :rtype: PyTango.DevVoid"""
        self.debug_stream("In " + self.get_name() + ".DbDeleteProperty()")
        # ----- PROTECTED REGION ID(DataBase.DbDeleteProperty) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbDeleteProperty

    # ------------------------------------------------------------------
    #    DbDeleteServer command:
    # ------------------------------------------------------------------
    def DbDeleteServer(self, argin):
        """Delete server from the database but dont delete device properties

        :param argin: Device server name
        :type: PyTango.DevString
        :return:
        :rtype: PyTango.DevVoid"""
        self.debug_stream("In " + self.get_name() + ".DbDeleteServer()")
        # ----- PROTECTED REGION ID(DataBase.DbDeleteServer) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbDeleteServer

    # ------------------------------------------------------------------
    #    DbDeleteServerInfo command:
    # ------------------------------------------------------------------
    def DbDeleteServerInfo(self, argin):
        """delete info related to a Tango devvice server process

        :param argin: Device server name
        :type: PyTango.DevString
        :return:
        :rtype: PyTango.DevVoid"""
        self.debug_stream("In " + self.get_name() + ".DbDeleteServerInfo()")
        # ----- PROTECTED REGION ID(DataBase.DbDeleteServerInfo) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbDeleteServerInfo

    # ------------------------------------------------------------------
    #    DbExportDevice command:
    # ------------------------------------------------------------------
    def DbExportDevice(self, argin):
        """Export a device to the database

        :param argin: Str[0] = Device name
        Str[1] = CORBA IOR
        Str[2] = Device server process host name
        Str[3] = Device server process PID or string ``null``
        Str[4] = Device server process version
        :type: PyTango.DevVarStringArray
        :return:
        :rtype: PyTango.DevVoid"""
        self.debug_stream("In " + self.get_name() + ".DbExportDevice()")
        # ----- PROTECTED REGION ID(DataBase.DbExportDevice) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbExportDevice

    # ------------------------------------------------------------------
    #    DbExportEvent command:
    # ------------------------------------------------------------------
    def DbExportEvent(self, argin):
        """Export Event channel to database

        :param argin: Str[0] = event channel name (or factory name)
        Str[1] = CORBA IOR
        Str[2] = Notifd host name
        Str[3] = Notifd pid
        Str[4] = Notifd version
        :type: PyTango.DevVarStringArray
        :return:
        :rtype: PyTango.DevVoid"""
        self.debug_stream("In " + self.get_name() + ".DbExportEvent()")
        # ----- PROTECTED REGION ID(DataBase.DbExportEvent) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbExportEvent

    # ------------------------------------------------------------------
    #    DbGetAliasDevice command:
    # ------------------------------------------------------------------
    def DbGetAliasDevice(self, argin):
        """Get device name from its alias.

        :param argin: Alias name
        :type: PyTango.DevString
        :return: Device name
        :rtype: PyTango.DevString"""
        self.debug_stream("In " + self.get_name() + ".DbGetAliasDevice()")
        argout = ""
        # ----- PROTECTED REGION ID(DataBase.DbGetAliasDevice) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetAliasDevice
        return argout

    # ------------------------------------------------------------------
    #    DbGetAttributeAlias command:
    # ------------------------------------------------------------------
    def DbGetAttributeAlias(self, argin):
        """Get the attribute name for the given alias.
        If alias not found in database, returns an empty string.

        :param argin: The attribute alias name
        :type: PyTango.DevString
        :return: The attribute name (device/attribute)
        :rtype: PyTango.DevString"""
        self.debug_stream("In " + self.get_name() + ".DbGetAttributeAlias()")
        argout = ""
        # ----- PROTECTED REGION ID(DataBase.DbGetAttributeAlias) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetAttributeAlias
        return argout

    # ------------------------------------------------------------------
    #    DbGetAttributeAliasList command:
    # ------------------------------------------------------------------
    def DbGetAttributeAliasList(self, argin):
        """Get attribute alias list for a specified filter

        :param argin: attribute alias filter string (eg: att*)
        :type: PyTango.DevString
        :return: attribute aliases
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetAttributeAliasList()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetAttributeAliasList) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetAttributeAliasList
        return argout

    # ------------------------------------------------------------------
    #    DbGetClassAttributeList command:
    # ------------------------------------------------------------------
    def DbGetClassAttributeList(self, argin):
        """Get attrilute list for a given Tango class with a specified filter

        :param argin: Str[0] = Tango class name
        Str[1] = Attribute name filter (eg: att*)
        :type: PyTango.DevVarStringArray
        :return: Str[0] = Class attribute name
        Str[n] = Class attribute name
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetClassAttributeList()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetClassAttributeList) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetClassAttributeList
        return argout

    # ------------------------------------------------------------------
    #    DbGetClassAttributeProperty command:
    # ------------------------------------------------------------------
    def DbGetClassAttributeProperty(self, argin):
        """Get Tango class property(ies) value

        :param argin: Str[0] = Tango class name
        Str[1] = Attribute name
        Str[n] = Attribute name
        :type: PyTango.DevVarStringArray
        :return: Str[0] = Tango class name
        Str[1] = Attribute property  number
        Str[2] = Attribute property 1 name
        Str[3] = Attribute property 1 value
        Str[n + 1] = Attribute property 2 name
        Str[n + 2] = Attribute property 2 value
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetClassAttributeProperty()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetClassAttributeProperty) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetClassAttributeProperty
        return argout

    # ------------------------------------------------------------------
    #    DbGetClassAttributeProperty2 command:
    # ------------------------------------------------------------------
    def DbGetClassAttributeProperty2(self, argin):
        """This command supports array property compared to the old command called
        DbGetClassAttributeProperty. The old command has not been deleted from the
        server for compatibility reasons.

        :param argin: Str[0] = Tango class name
        Str[1] = Attribute name
        Str[n] = Attribute name
        :type: PyTango.DevVarStringArray
        :return: Str[0] = Tango class name
        Str[1] = Attribute property  number
        Str[2] = Attribute property 1 name
        Str[3] = Attribute property 1 value number (array case)
        Str[4] = Attribute property 1 value
        Str[n] = Attribute property 1 value (array case)
        Str[n + 1] = Attribute property 2 name
        Str[n + 2] = Attribute property 2 value number (array case)
        Str[n + 3] = Attribute property 2 value
        Str[n + m] = Attribute property 2 value (array case)
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetClassAttributeProperty2()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetClassAttributeProperty2) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetClassAttributeProperty2
        return argout

    # ------------------------------------------------------------------
    #    DbGetClassAttributePropertyHist command:
    # ------------------------------------------------------------------
    def DbGetClassAttributePropertyHist(self, argin):
        """Retrieve Tango class attribute property history

        :param argin: Str[0] = Tango class
        Str[1] = Attribute name
        Str[2] = Property name
        :type: PyTango.DevVarStringArray
        :return: Str[0] = Attribute name
        Str[1] = Property name
        Str[2] = date
        Str[3] = Property value number (array case)
        Str[4] = Property value 1
        Str[n] = Property value n
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream(
            "In " + self.get_name() + ".DbGetClassAttributePropertyHist()"
        )
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetClassAttributePropertyHist) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetClassAttributePropertyHist
        return argout

    # ------------------------------------------------------------------
    #    DbGetClassForDevice command:
    # ------------------------------------------------------------------
    def DbGetClassForDevice(self, argin):
        """Get Tango class for the specified device.

        :param argin: Device name
        :type: PyTango.DevString
        :return: Device Tango class
        :rtype: PyTango.DevString"""
        self.debug_stream("In " + self.get_name() + ".DbGetClassForDevice()")
        argout = ""
        # ----- PROTECTED REGION ID(DataBase.DbGetClassForDevice) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetClassForDevice
        return argout

    # ------------------------------------------------------------------
    #    DbGetClassInheritanceForDevice command:
    # ------------------------------------------------------------------
    def DbGetClassInheritanceForDevice(self, argin):
        """Get class inheritance for the specified device.

        :param argin: Device name
        :type: PyTango.DevString
        :return: Classes off the specified device.
        [0] - is the class of the device.
        [1] - is the class from the device class is inherited.
        ........and so on
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetClassInheritanceForDevice()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetClassInheritanceForDevice) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetClassInheritanceForDevice
        return argout

    # ------------------------------------------------------------------
    #    DbGetClassList command:
    # ------------------------------------------------------------------
    def DbGetClassList(self, argin):
        """Get Tango class list with a specified filter

        :param argin: Filter
        :type: PyTango.DevString
        :return: Class list
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetClassList()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetClassList) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetClassList
        return argout

    # ------------------------------------------------------------------
    #    DbGetClassProperty command:
    # ------------------------------------------------------------------
    def DbGetClassProperty(self, argin):
        """

        :param argin: Str[0] = Tango class
        Str[1] = Property name
        Str[2] = Property name
        :type: PyTango.DevVarStringArray
        :return: Str[0] = Tango class
        Str[1] = Property number
        Str[2] = Property name
        Str[3] = Property value number (array case)
        Str[4] = Property value
        Str[n] = Propery value (array case)
        ....
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetClassProperty()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetClassProperty) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetClassProperty
        return argout

    # ------------------------------------------------------------------
    #    DbGetClassPropertyHist command:
    # ------------------------------------------------------------------
    def DbGetClassPropertyHist(self, argin):
        """Retrieve Tango class property history

        :param argin: Str[0] = Tango class
        Str[1] = Property name
        :type: PyTango.DevVarStringArray
        :return: Str[0] = Property name
        Str[1] = date
        Str[2] = Property value number (array case)
        Str[3] = Property value 1
        Str[n] = Property value n
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetClassPropertyHist()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetClassPropertyHist) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetClassPropertyHist
        return argout

    # ------------------------------------------------------------------
    #    DbGetClassPropertyList command:
    # ------------------------------------------------------------------
    def DbGetClassPropertyList(self, argin):
        """Get property list for a given Tango class with a specified filter

        :param argin: The filter
        :type: PyTango.DevString
        :return: Property name list
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetClassPropertyList()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetClassPropertyList) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetClassPropertyList
        return argout

    # ------------------------------------------------------------------
    #    DbGetDeviceAlias command:
    # ------------------------------------------------------------------
    def DbGetDeviceAlias(self, argin):
        """Return alias for device name if found.

        :param argin: The device name
        :type: PyTango.DevString
        :return: The alias found
        :rtype: PyTango.DevString"""
        self.debug_stream("In " + self.get_name() + ".DbGetDeviceAlias()")
        argout = ""
        # ----- PROTECTED REGION ID(DataBase.DbGetDeviceAlias) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetDeviceAlias
        return argout

    # ------------------------------------------------------------------
    #    DbGetDeviceAliasList command:
    # ------------------------------------------------------------------
    def DbGetDeviceAliasList(self, argin):
        """Get device alias name with a specific filter

        :param argin: The filter
        :type: PyTango.DevString
        :return: Device alias list
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetDeviceAliasList()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetDeviceAliasList) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetDeviceAliasList
        return argout

    # ------------------------------------------------------------------
    #    DbGetDeviceAttributeList command:
    # ------------------------------------------------------------------
    def DbGetDeviceAttributeList(self, argin):
        """Return list of attributes matching the wildcard
         for the specified device

        :param argin: Str[0] = Device name
        Str[1] = Wildcard
        :type: PyTango.DevVarStringArray
        :return: attribute name list
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetDeviceAttributeList()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetDeviceAttributeList) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetDeviceAttributeList
        return argout

    # ------------------------------------------------------------------
    #    DbGetDeviceAttributeProperty command:
    # ------------------------------------------------------------------
    def DbGetDeviceAttributeProperty(self, argin):
        """Get device attribute property(ies) value

        :param argin: Str[0] = Device name
        Str[1] = Attribute name
        Str[n] = Attribute name
        :type: PyTango.DevVarStringArray
        :return: Str[0] = Device name
        Str[1] = Attribute property  number
        Str[2] = Attribute property 1 name
        Str[3] = Attribute property 1 value
        Str[n + 1] = Attribute property 2 name
        Str[n + 2] = Attribute property 2 value
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetDeviceAttributeProperty()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetDeviceAttributeProperty) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetDeviceAttributeProperty
        return argout

    # ------------------------------------------------------------------
    #    DbGetDeviceAttributeProperty2 command:
    # ------------------------------------------------------------------
    def DbGetDeviceAttributeProperty2(self, argin):
        """Retrieve device attribute properties. This command has the possibility to retrieve
        device attribute properties which are arrays. It is not possible with the old
        DbGetDeviceAttributeProperty command. Nevertheless, the old command has not been
        deleted for compatibility reason

        :param argin: Str[0] = Device name
        Str[1] = Attribute name
        Str[n] = Attribute name
        :type: PyTango.DevVarStringArray
        :return: Str[0] = Device name
        Str[1] = Attribute property  number
        Str[2] = Attribute property 1 name
        Str[3] = Attribute property 1 value number (array case)
        Str[4] = Attribute property 1 value
        Str[n] = Attribute property 1 value (array case)
        Str[n + 1] = Attribute property 2 name
        Str[n + 2] = Attribute property 2 value number (array case)
        Str[n + 3] = Attribute property 2 value
        Str[n + m] = Attribute property 2 value (array case)
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetDeviceAttributeProperty2()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetDeviceAttributeProperty2) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetDeviceAttributeProperty2
        return argout

    # ------------------------------------------------------------------
    #    DbGetDeviceAttributePropertyHist command:
    # ------------------------------------------------------------------
    def DbGetDeviceAttributePropertyHist(self, argin):
        """Retrieve device attribute property history

        :param argin: Str[0] = Device name
        Str[1] = Attribute name
        Str[2] = Property name
        :type: PyTango.DevVarStringArray
        :return: Str[0] = Attribute name
        Str[1] = Property name
        Str[2] = date
        Str[3] = Property value number (array case)
        Str[4] = Property value 1
        Str[n] = Property value n
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream(
            "In " + self.get_name() + ".DbGetDeviceAttributePropertyHist()"
        )
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetDeviceAttributePropertyHist) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetDeviceAttributePropertyHist
        return argout

    # ------------------------------------------------------------------
    #    DbGetDeviceClassList command:
    # ------------------------------------------------------------------
    def DbGetDeviceClassList(self, argin):
        """Get Tango classes/device list embedded in a specific device server

        :param argin: Device server process name
        :type: PyTango.DevString
        :return: Str[0] = Device name
        Str[1] = Tango class
        Str[n] = Device name
        Str[n + 1] = Tango class
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetDeviceClassList()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetDeviceClassList) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetDeviceClassList
        return argout

    # ------------------------------------------------------------------
    #    DbGetDeviceDomainList command:
    # ------------------------------------------------------------------
    def DbGetDeviceDomainList(self, argin):
        """Get list of device domain name matching the specified

        :param argin: The wildcard
        :type: PyTango.DevString
        :return: Device name domain list
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetDeviceDomainList()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetDeviceDomainList) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetDeviceDomainList
        return argout

    # ------------------------------------------------------------------
    #    DbGetDeviceExportedList command:
    # ------------------------------------------------------------------
    def DbGetDeviceExportedList(self, argin):
        """Get a list of exported devices whose names satisfy the filter (wildcard is

        :param argin: filter
        :type: PyTango.DevString
        :return: list of exported devices
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetDeviceExportedList()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetDeviceExportedList) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetDeviceExportedList
        return argout

    # ------------------------------------------------------------------
    #    DbGetDeviceFamilyList command:
    # ------------------------------------------------------------------
    def DbGetDeviceFamilyList(self, argin):
        """Get a list of device name families for device name matching the
        specified wildcard

        :param argin: The wildcard
        :type: PyTango.DevString
        :return: Family list
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetDeviceFamilyList()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetDeviceFamilyList) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetDeviceFamilyList
        return argout

    # ------------------------------------------------------------------
    #    DbGetDeviceInfo command:
    # ------------------------------------------------------------------
    def DbGetDeviceInfo(self, argin):
        """Returns info from DbImportDevice and started/stopped dates.

        :param argin: Device name
        :type: PyTango.DevString
        :return: Str[0] = Device name
        Str[1] = CORBA IOR
        Str[2] = Device version
        Str[3] = Device Server name
        Str[4] = Device Server process host name
        Str[5] = Started date (or ? if not set)
        Str[6] = Stopped date (or ? if not set)
        Str[7] = Device class

        Lg[0] = Device exported flag
        Lg[1] = Device Server process PID (or -1 if not set)
        :rtype: PyTango.DevVarLongStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetDeviceInfo()")
        argout = [0], [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetDeviceInfo) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetDeviceInfo
        return argout

    # ------------------------------------------------------------------
    #    DbGetDeviceList command:
    # ------------------------------------------------------------------
    def DbGetDeviceList(self, argin):
        """Get a list of devices for specified server and class.

        :param argin: argin[0] : server name
        argin[1] : class name
        :type: PyTango.DevVarStringArray
        :return: The list of devices for specified server and class.
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetDeviceList()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetDeviceList) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetDeviceList
        return argout

    # ------------------------------------------------------------------
    #    DbGetDeviceWideList command:
    # ------------------------------------------------------------------
    def DbGetDeviceWideList(self, argin):
        """Get a list of devices whose names satisfy the filter.

        :param argin: filter
        :type: PyTango.DevString
        :return: list of exported devices
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetDeviceWideList()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetDeviceWideList) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetDeviceWideList
        return argout

    # ------------------------------------------------------------------
    #    DbGetDeviceMemberList command:
    # ------------------------------------------------------------------
    def DbGetDeviceMemberList(self, argin):
        """Get a list of device name members for device name matching the
        specified filter

        :param argin: The filter
        :type: PyTango.DevString
        :return: Device names member list
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetDeviceMemberList()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetDeviceMemberList) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetDeviceMemberList
        return argout

    # ------------------------------------------------------------------
    #    DbGetDeviceProperty command:
    # ------------------------------------------------------------------
    def DbGetDeviceProperty(self, argin):
        """

        :param argin: Str[0] = Device name
        Str[1] = Property name
        Str[n] = Property name
        :type: PyTango.DevVarStringArray
        :return: Str[0] = Device name
        Str[1] = Property number
        Str[2] = Property name
        Str[3] = Property value number (array case)
        Str[4] = Property value 1
        Str[n] = Property value n (array case)
        Str[n + 1] = Property name
        Str[n + 2] = Property value number (array case)
        Str[n + 3] = Property value 1
        Str[n + m] = Property value m
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetDeviceProperty()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetDeviceProperty) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetDeviceProperty
        return argout

    # ------------------------------------------------------------------
    #    DbGetDevicePropertyHist command:
    # ------------------------------------------------------------------
    def DbGetDevicePropertyHist(self, argin):
        """Retrieve device  property history

        :param argin: Str[0] = Device name
        Str[2] = Property name
        :type: PyTango.DevVarStringArray
        :return: Str[0] = Property name
        Str[1] = date
        Str[2] = Property value number (array case)
        Str[3] = Property value 1
        Str[n] = Property value n
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetDevicePropertyHist()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetDevicePropertyHist) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetDevicePropertyHist
        return argout

    # ------------------------------------------------------------------
    #    DbGetDevicePropertyList command:
    # ------------------------------------------------------------------
    def DbGetDevicePropertyList(self, argin):
        """Get property list belonging to the specified device and with
        name matching the specified filter

        :param argin: Str[0] = device name
        Str[1] = Filter
        :type: PyTango.DevVarStringArray
        :return: Property name list
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetDevicePropertyList()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetDevicePropertyList) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetDevicePropertyList
        return argout

    # ------------------------------------------------------------------
    #    DbGetDeviceServerClassList command:
    # ------------------------------------------------------------------
    def DbGetDeviceServerClassList(self, argin):
        """Get list of Tango classes for a device server

        :param argin: device server process name
        :type: PyTango.DevString
        :return: list of classes for this device server
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetDeviceServerClassList()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetDeviceServerClassList) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetDeviceServerClassList
        return argout

    # ------------------------------------------------------------------
    #    DbGetExportdDeviceListForClass command:
    # ------------------------------------------------------------------
    def DbGetExportdDeviceListForClass(self, argin):
        """Query the database for device exported for the specified class.

        :param argin: Class name
        :type: PyTango.DevString
        :return: Device exported list
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetExportdDeviceListForClass()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetExportdDeviceListForClass) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetExportdDeviceListForClass
        return argout

    # ------------------------------------------------------------------
    #    DbGetHostList command:
    # ------------------------------------------------------------------
    def DbGetHostList(self, argin):
        """Get host list with name matching the specified filter

        :param argin: The filter
        :type: PyTango.DevString
        :return: Host name list
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetHostList()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetHostList) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetHostList
        return argout

    # ------------------------------------------------------------------
    #    DbGetHostServerList command:
    # ------------------------------------------------------------------
    def DbGetHostServerList(self, argin):
        """Get list of device server process name running on host with name matching
        the specified filter

        :param argin: The filter
        :type: PyTango.DevString
        :return: Device server process name list
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetHostServerList()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetHostServerList) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetHostServerList
        return argout

    # ------------------------------------------------------------------
    #    DbGetHostServersInfo command:
    # ------------------------------------------------------------------
    def DbGetHostServersInfo(self, argin):
        """Get info about all servers running on specified host, name, mode and level

        :param argin: Host name
        :type: PyTango.DevString
        :return: Server info for all servers running on specified host
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetHostServersInfo()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetHostServersInfo) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetHostServersInfo
        return argout

    # ------------------------------------------------------------------
    #    DbGetInstanceNameList command:
    # ------------------------------------------------------------------
    def DbGetInstanceNameList(self, argin):
        """Returns the instance names found for specified server.

        :param argin: Server name
        :type: PyTango.DevString
        :return: The instance names found for specified server.
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetInstanceNameList()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetInstanceNameList) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetInstanceNameList
        return argout

    # ------------------------------------------------------------------
    #    DbGetObjectList command:
    # ------------------------------------------------------------------
    def DbGetObjectList(self, argin):
        """Get list of free object defined in database with name
        matching the specified filter

        :param argin: The filter
        :type: PyTango.DevString
        :return: Object name list
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetObjectList()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetObjectList) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetObjectList
        return argout

    # ------------------------------------------------------------------
    #    DbGetProperty command:
    # ------------------------------------------------------------------
    def DbGetProperty(self, argin):
        """Get free object property

        :param argin: Str[0] = Object name
        Str[1] = Property name
        Str[n] = Property name
        :type: PyTango.DevVarStringArray
        :return: Str[0] = Object name
        Str[1] = Property number
        Str[2] = Property name
        Str[3] = Property value number (array case)
        Str[4] = Property value 1
        Str[n] = Property value n (array case)
        Str[n + 1] = Property name
        Str[n + 2] = Property value number (array case)
        Str[n + 3] = Property value 1
        Str[n + m] = Property value m
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetProperty()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetProperty) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetProperty
        return argout

    # ------------------------------------------------------------------
    #    DbGetPropertyHist command:
    # ------------------------------------------------------------------
    def DbGetPropertyHist(self, argin):
        """Retrieve object  property history

        :param argin: Str[0] = Object name
        Str[2] = Property name
        :type: PyTango.DevVarStringArray
        :return: Str[0] = Property name
        Str[1] = date
        Str[2] = Property value number (array case)
        Str[3] = Property value 1
        Str[n] = Property value n
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetPropertyHist()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetPropertyHist) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetPropertyHist
        return argout

    # ------------------------------------------------------------------
    #    DbGetPropertyList command:
    # ------------------------------------------------------------------
    def DbGetPropertyList(self, argin):
        """Get list of property defined for a free object and matching the
        specified filter

        :param argin: Str[0] = Object name
        Str[1] = filter
        :type: PyTango.DevVarStringArray
        :return: Property name list
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetPropertyList()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetPropertyList) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetPropertyList
        return argout

    # ------------------------------------------------------------------
    #    DbGetServerInfo command:
    # ------------------------------------------------------------------
    def DbGetServerInfo(self, argin):
        """Get info about host, mode and level for specified server

        :param argin: server name
        :type: PyTango.DevString
        :return: server info
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetServerInfo()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetServerInfo) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetServerInfo
        return argout

    # ------------------------------------------------------------------
    #    DbGetServerList command:
    # ------------------------------------------------------------------
    def DbGetServerList(self, argin):
        """Get list of device server process defined in database
        with name matching the specified filter

        :param argin: The filter
        :type: PyTango.DevString
        :return: Device server process name list
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetServerList()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetServerList) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetServerList
        return argout

    # ------------------------------------------------------------------
    #    DbGetServerNameList command:
    # ------------------------------------------------------------------
    def DbGetServerNameList(self, argin):
        """Returns the list of server names found for the wildcard specified.
        It returns only the server executable name without instance name as DbGetServerList.

        :param argin: wildcard for server names.
        :type: PyTango.DevString
        :return: server names found.
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetServerNameList()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetServerNameList) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetServerNameList
        return argout

    # ------------------------------------------------------------------
    #    DbImportDevice command:
    # ------------------------------------------------------------------
    def DbImportDevice(self, argin):
        """Import a device from the database

        :param argin: Device name (or alias)
        :type: PyTango.DevString
        :return: Str[0] = device name
        Str[1] = CORBA IOR
        Str[2] = device version
        Str[3] = device server process name
        Str[4] = host name
        Str[5] = Tango class name

        Lg[0] = Exported flag
        Lg[1] = Device server process PID
        :rtype: PyTango.DevVarLongStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbImportDevice()")
        argout = [0], [""]
        # ----- PROTECTED REGION ID(DataBase.DbImportDevice) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbImportDevice
        return argout

    # ------------------------------------------------------------------
    #    DbImportEvent command:
    # ------------------------------------------------------------------
    def DbImportEvent(self, argin):
        """Get event channel info from database

        :param argin: name of event channel or factory
        :type: PyTango.DevString
        :return: export information e.g. IOR
        :rtype: PyTango.DevVarLongStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbImportEvent()")
        argout = [0], [""]
        # ----- PROTECTED REGION ID(DataBase.DbImportEvent) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbImportEvent
        return argout

    # ------------------------------------------------------------------
    #    DbInfo command:
    # ------------------------------------------------------------------
    def DbInfo(self):
        """Get miscellaneous numbers on information
        stored in database

        :param :
        :type: PyTango.DevVoid
        :return: Miscellaneous info like:
        - Device defined in database
        - Device marked as exported in database
        - Device server process defined in database
        - Device server process marked as exported in database
        - Device properties defined in database
        - Class properties defined in database
        - Device attribute properties defined in database
        - Class attribute properties defined in database
        - Object properties defined in database
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbInfo()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbInfo) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbInfo
        return argout

    # ------------------------------------------------------------------
    #    DbPutAttributeAlias command:
    # ------------------------------------------------------------------
    def DbPutAttributeAlias(self, argin):
        """Define an alias for an attribute

        :param argin: Str[0] = attribute name
        Str[1] = attribute alias
        :type: PyTango.DevVarStringArray
        :return:
        :rtype: PyTango.DevVoid"""
        self.debug_stream("In " + self.get_name() + ".DbPutAttributeAlias()")
        # ----- PROTECTED REGION ID(DataBase.DbPutAttributeAlias) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbPutAttributeAlias

    # ------------------------------------------------------------------
    #    DbPutClassAttributeProperty command:
    # ------------------------------------------------------------------
    def DbPutClassAttributeProperty(self, argin):
        """Create/Update class attribute property(ies) in database

        :param argin: Str[0] = Tango class name
        Str[1] = Attribute number
        Str[2] = Attribute name
        Str[3] = Property number
        Str[4] = Property name
        Str[5] = Property value
        .....
        :type: PyTango.DevVarStringArray
        :return:
        :rtype: PyTango.DevVoid"""
        self.debug_stream("In " + self.get_name() + ".DbPutClassAttributeProperty()")
        # ----- PROTECTED REGION ID(DataBase.DbPutClassAttributeProperty) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbPutClassAttributeProperty

    # ------------------------------------------------------------------
    #    DbPutClassAttributeProperty2 command:
    # ------------------------------------------------------------------
    def DbPutClassAttributeProperty2(self, argin):
        """This command adds support for array properties compared to the previous one
        called DbPutClassAttributeProperty. The old comman is still there for compatibility reason

        :param argin: Str[0] = Tango class name
        Str[1] = Attribute number
        Str[2] = Attribute name
        Str[3] = Property number
        Str[4] = Property name
        Str[5] = Property value number (array case)
        Str[5] = Property value 1
        Str[n] = Property value n (array case)
        .....
        :type: PyTango.DevVarStringArray
        :return:
        :rtype: PyTango.DevVoid"""
        self.debug_stream("In " + self.get_name() + ".DbPutClassAttributeProperty2()")
        # ----- PROTECTED REGION ID(DataBase.DbPutClassAttributeProperty2) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbPutClassAttributeProperty2

    # ------------------------------------------------------------------
    #    DbPutClassProperty command:
    # ------------------------------------------------------------------
    def DbPutClassProperty(self, argin):
        """Create / Update class property(ies)

        :param argin: Str[0] = Tango class name
        Str[1] = Property number
        Str[2] = Property name
        Str[3] = Property value number
        Str[4] = Property value 1
        Str[n] = Property value n
        ....
        :type: PyTango.DevVarStringArray
        :return:
        :rtype: PyTango.DevVoid"""
        self.debug_stream("In " + self.get_name() + ".DbPutClassProperty()")
        # ----- PROTECTED REGION ID(DataBase.DbPutClassProperty) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbPutClassProperty

    # ------------------------------------------------------------------
    #    DbPutDeviceAlias command:
    # ------------------------------------------------------------------
    def DbPutDeviceAlias(self, argin):
        """Define alias for  a given device name

        :param argin: Str[0] = device name
        Str[1] = alias name
        :type: PyTango.DevVarStringArray
        :return:
        :rtype: PyTango.DevVoid"""
        self.debug_stream("In " + self.get_name() + ".DbPutDeviceAlias()")
        # ----- PROTECTED REGION ID(DataBase.DbPutDeviceAlias) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbPutDeviceAlias

    # ------------------------------------------------------------------
    #    DbPutDeviceAttributeProperty command:
    # ------------------------------------------------------------------
    def DbPutDeviceAttributeProperty(self, argin):
        """Create/Update device attribute property(ies) in database

        :param argin: Str[0] = Device name
        Str[1] = Attribute number
        Str[2] = Attribute name
        Str[3] = Property number
        Str[4] = Property name
        Str[5] = Property value
        .....
        :type: PyTango.DevVarStringArray
        :return:
        :rtype: PyTango.DevVoid"""
        self.debug_stream("In " + self.get_name() + ".DbPutDeviceAttributeProperty()")
        # ----- PROTECTED REGION ID(DataBase.DbPutDeviceAttributeProperty) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbPutDeviceAttributeProperty

    # ------------------------------------------------------------------
    #    DbPutDeviceAttributeProperty2 command:
    # ------------------------------------------------------------------
    def DbPutDeviceAttributeProperty2(self, argin):
        """Put device attribute property. This command adds the possibility to have attribute property
        which are arrays. Not possible with the old DbPutDeviceAttributeProperty command.
        This old command is not deleted for compatibility reasons.

        :param argin: Str[0] = Device name
        Str[1] = Attribute number
        Str[2] = Attribute name
        Str[3] = Property number
        Str[4] = Property name
        Str[5] = Property value number (array case)
        Str[5] = Property value 1
        Str[n] = Property value n (array case)
        .....
        :type: PyTango.DevVarStringArray
        :return:
        :rtype: PyTango.DevVoid"""
        self.debug_stream("In " + self.get_name() + ".DbPutDeviceAttributeProperty2()")
        # ----- PROTECTED REGION ID(DataBase.DbPutDeviceAttributeProperty2) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbPutDeviceAttributeProperty2

    # ------------------------------------------------------------------
    #    DbPutDeviceProperty command:
    # ------------------------------------------------------------------
    def DbPutDeviceProperty(self, argin):
        """Create / Update device property(ies)

        :param argin: Str[0] = Tango device name
        Str[1] = Property number
        Str[2] = Property name
        Str[3] = Property value number
        Str[4] = Property value 1
        Str[n] = Property value n
        ....
        :type: PyTango.DevVarStringArray
        :return:
        :rtype: PyTango.DevVoid"""
        self.debug_stream("In " + self.get_name() + ".DbPutDeviceProperty()")
        # ----- PROTECTED REGION ID(DataBase.DbPutDeviceProperty) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbPutDeviceProperty

    # ------------------------------------------------------------------
    #    DbPutProperty command:
    # ------------------------------------------------------------------
    def DbPutProperty(self, argin):
        """Create / Update free object property(ies)

        :param argin: Str[0] = Object name
        Str[1] = Property number
        Str[2] = Property name
        Str[3] = Property value number
        Str[4] = Property value 1
        Str[n] = Property value n
        ....
        :type: PyTango.DevVarStringArray
        :return:
        :rtype: PyTango.DevVoid"""
        self.debug_stream("In " + self.get_name() + ".DbPutProperty()")
        # ----- PROTECTED REGION ID(DataBase.DbPutProperty) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbPutProperty

    # ------------------------------------------------------------------
    #    DbPutServerInfo command:
    # ------------------------------------------------------------------
    def DbPutServerInfo(self, argin):
        """Update server info including host, mode and level

        :param argin: server info
        :type: PyTango.DevVarStringArray
        :return:
        :rtype: PyTango.DevVoid"""
        self.debug_stream("In " + self.get_name() + ".DbPutServerInfo()")
        # ----- PROTECTED REGION ID(DataBase.DbPutServerInfo) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbPutServerInfo

    # ------------------------------------------------------------------
    #    DbUnExportDevice command:
    # ------------------------------------------------------------------
    def DbUnExportDevice(self, argin):
        """Mark a device as non exported in database

        :param argin: Device name
        :type: PyTango.DevString
        :return:
        :rtype: PyTango.DevVoid"""
        self.debug_stream("In " + self.get_name() + ".DbUnExportDevice()")
        # ----- PROTECTED REGION ID(DataBase.DbUnExportDevice) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbUnExportDevice

    # ------------------------------------------------------------------
    #    DbUnExportEvent command:
    # ------------------------------------------------------------------
    def DbUnExportEvent(self, argin):
        """Mark one event channel as non exported in database

        :param argin: name of event channel or factory to unexport
        :type: PyTango.DevString
        :return: none
        :rtype: PyTango.DevVoid"""
        self.debug_stream("In " + self.get_name() + ".DbUnExportEvent()")
        # ----- PROTECTED REGION ID(DataBase.DbUnExportEvent) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbUnExportEvent

    # ------------------------------------------------------------------
    #    DbUnExportServer command:
    # ------------------------------------------------------------------
    def DbUnExportServer(self, argin):
        """Mark all devices belonging to a specified device server
        process as non exported

        :param argin: Device server name (executable/instance)
        :type: PyTango.DevString
        :return:
        :rtype: PyTango.DevVoid"""
        self.debug_stream("In " + self.get_name() + ".DbUnExportServer()")
        # ----- PROTECTED REGION ID(DataBase.DbUnExportServer) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbUnExportServer

    # ------------------------------------------------------------------
    #    ResetTimingValues command:
    # ------------------------------------------------------------------
    def ResetTimingValues(self):
        """Reset the timing attribute values.

        :param :
        :type: PyTango.DevVoid
        :return:
        :rtype: PyTango.DevVoid"""
        self.debug_stream("In " + self.get_name() + ".ResetTimingValues()")
        # ----- PROTECTED REGION ID(DataBase.ResetTimingValues) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.ResetTimingValues

    # ------------------------------------------------------------------
    #    DbGetDataForServerCache command:
    # ------------------------------------------------------------------
    def DbGetDataForServerCache(self, argin):
        """This command returns all the data needed by a device server process during its
        startup sequence. The aim of this command is to minimize database access during
        device server startup sequence.

        :param argin: Elt[0] = DS name (exec_name/inst_name), Elt[1] = Host name
        :type: PyTango.DevVarStringArray
        :return: All the data needed by the device server during its startup sequence. Precise list depend on the device server
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetDataForServerCache()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetDataForServerCache) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetDataForServerCache
        return argout

    # ------------------------------------------------------------------
    #    DbDeleteAllDeviceAttributeProperty command:
    # ------------------------------------------------------------------
    def DbDeleteAllDeviceAttributeProperty(self, argin):
        """Delete all attribute properties for the specified device attribute(s)

        :param argin: str[0] = device name
        Str[1]...str[n] = attribute name(s)
        :type: PyTango.DevVarStringArray
        :return:
        :rtype: PyTango.DevVoid"""
        self.debug_stream(
            "In " + self.get_name() + ".DbDeleteAllDeviceAttributeProperty()"
        )
        # ----- PROTECTED REGION ID(DataBase.DbDeleteAllDeviceAttributeProperty) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbDeleteAllDeviceAttributeProperty

    # ------------------------------------------------------------------
    #    DbMySqlSelect command:
    # ------------------------------------------------------------------
    def DbMySqlSelect(self, argin):
        """This is a very low level command.
        It executes the specified  SELECT command on TANGO database and returns its result without filter.

        :param argin: MySql Select command
        :type: PyTango.DevString
        :return: MySql Select command result
         - svalues : select results
         - lvalue[n] : =0 if svalue[n] is null else =1
         (last lvalue -1) is number of rows, (last lvalue) is number of fields
        :rtype: PyTango.DevVarLongStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbMySqlSelect()")
        argout = [0], [""]
        # ----- PROTECTED REGION ID(DataBase.DbMySqlSelect) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbMySqlSelect
        return argout

    # ------------------------------------------------------------------
    #    DbGetCSDbServerList command:
    # ------------------------------------------------------------------
    def DbGetCSDbServerList(self):
        """Get a list of host:port for all database server defined in the control system

        :param :
        :type: PyTango.DevVoid
        :return: List of host:port with one element for each database server
        :rtype: PyTango.DevVarStringArray"""
        self.debug_stream("In " + self.get_name() + ".DbGetCSDbServerList()")
        argout = [""]
        # ----- PROTECTED REGION ID(DataBase.DbGetCSDbServerList) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetCSDbServerList
        return argout

    # ------------------------------------------------------------------
    #    DbGetAttributeAlias2 command:
    # ------------------------------------------------------------------
    def DbGetAttributeAlias2(self, argin):
        """Get the attribute alias from the attribute name.
        Returns one empty string if nothing found in database

        :param argin: The attribute name (dev_name/att_name)
        :type: PyTango.DevString
        :return: The attribute alias name (or empty string)
        :rtype: PyTango.DevString"""
        self.debug_stream("In " + self.get_name() + ".DbGetAttributeAlias2()")
        argout = ""
        # ----- PROTECTED REGION ID(DataBase.DbGetAttributeAlias2) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetAttributeAlias2
        return argout

    # ------------------------------------------------------------------
    #    DbGetAliasAttribute command:
    # ------------------------------------------------------------------
    def DbGetAliasAttribute(self, argin):
        """Get the attribute name from the given alias.
        If the given alias is not found in database, returns an empty string

        :param argin: The attribute alias
        :type: PyTango.DevString
        :return: The attribute name (dev_name/att_name)
        :rtype: PyTango.DevString"""
        self.debug_stream("In " + self.get_name() + ".DbGetAliasAttribute()")
        argout = ""
        # ----- PROTECTED REGION ID(DataBase.DbGetAliasAttribute) ENABLED START -----#

        # ----- PROTECTED REGION END -----#	//	DataBase.DbGetAliasAttribute
        return argout


# ==================================================================
#
#    DataBaseClass class definition
#
# ==================================================================
class DataBaseClass(PyTango.DeviceClass):
    #    Class Properties
    class_property_list = {}

    #    Device Properties
    device_property_list = {}

    #    Command definitions
    cmd_list = {
        "DbAddDevice": [
            [
                PyTango.DevVarStringArray,
                "Str[0] = Full device server process name\nStr[1] = Device name\nStr[2] = Tango class name",
            ],
            [PyTango.DevVoid, "none"],
        ],
        "DbAddServer": [
            [
                PyTango.DevVarStringArray,
                "Str[0] = Full device server name\nStr[1] = Device(s) name\nStr[2] = Tango class name\nStr[n] = Device name\nStr[n + 1] = Tango class name",
            ],
            [PyTango.DevVoid, "none"],
        ],
        "DbDeleteAttributeAlias": [
            [PyTango.DevString, "Attriibute alias name."],
            [PyTango.DevVoid, "none"],
        ],
        "DbDeleteClassAttribute": [
            [
                PyTango.DevVarStringArray,
                "Str[0] = Tango class name\nStr[1] = Attribute name",
            ],
            [PyTango.DevVoid, "none"],
        ],
        "DbDeleteClassAttributeProperty": [
            [
                PyTango.DevVarStringArray,
                "Str[0] = Tango class name\nStr[1] = Attribute name\nStr[2] = Property name\nStr[n] = Property name",
            ],
            [PyTango.DevVoid, "none"],
        ],
        "DbDeleteClassProperty": [
            [
                PyTango.DevVarStringArray,
                "Str[0] = Tango class name\nStr[1] = Property name\nStr[n] = Property name",
            ],
            [PyTango.DevVoid, "none"],
        ],
        "DbDeleteDevice": [
            [PyTango.DevString, "device name"],
            [PyTango.DevVoid, "none"],
        ],
        "DbDeleteDeviceAlias": [
            [PyTango.DevString, "device alias name"],
            [PyTango.DevVoid, "none"],
        ],
        "DbDeleteDeviceAttribute": [
            [
                PyTango.DevVarStringArray,
                "Str[0] = Device name\nStr[1] = Attribute name",
            ],
            [PyTango.DevVoid, "none"],
        ],
        "DbDeleteDeviceAttributeProperty": [
            [
                PyTango.DevVarStringArray,
                "Str[0] = Device name\nStr[1] = Attribute name\nStr[2] = Property name\nStr[n] = Property name",
            ],
            [PyTango.DevVoid, "none"],
        ],
        "DbDeleteDeviceProperty": [
            [
                PyTango.DevVarStringArray,
                "Str[0] = Device name\nStr[1] = Property name\nStr[n] = Property name",
            ],
            [PyTango.DevVoid, "none"],
        ],
        "DbDeleteProperty": [
            [
                PyTango.DevVarStringArray,
                "Str[0]  = Object name\nStr[1] = Property name\nStr[n] = Property name",
            ],
            [PyTango.DevVoid, "none"],
        ],
        "DbDeleteServer": [
            [PyTango.DevString, "Device server name"],
            [PyTango.DevVoid, "none"],
        ],
        "DbDeleteServerInfo": [
            [PyTango.DevString, "Device server name"],
            [PyTango.DevVoid, "none"],
        ],
        "DbExportDevice": [
            [
                PyTango.DevVarStringArray,
                "Str[0] = Device name\nStr[1] = CORBA IOR\nStr[2] = Device server process host name\nStr[3] = Device server process PID or string ``null``\nStr[4] = Device server process version",
            ],
            [PyTango.DevVoid, "none"],
        ],
        "DbExportEvent": [
            [
                PyTango.DevVarStringArray,
                "Str[0] = event channel name (or factory name)\nStr[1] = CORBA IOR\nStr[2] = Notifd host name\nStr[3] = Notifd pid\nStr[4] = Notifd version",
            ],
            [PyTango.DevVoid, "none"],
        ],
        "DbGetAliasDevice": [
            [PyTango.DevString, "Alias name"],
            [PyTango.DevString, "Device name"],
        ],
        "DbGetAttributeAlias": [
            [PyTango.DevString, "The attribute alias name"],
            [PyTango.DevString, "The attribute name (device/attribute)"],
        ],
        "DbGetAttributeAliasList": [
            [PyTango.DevString, "attribute alias filter string (eg: att*)"],
            [PyTango.DevVarStringArray, "attribute aliases"],
        ],
        "DbGetClassAttributeList": [
            [
                PyTango.DevVarStringArray,
                "Str[0] = Tango class name\nStr[1] = Attribute name filter (eg: att*)",
            ],
            [
                PyTango.DevVarStringArray,
                "Str[0] = Class attribute name\nStr[n] = Class attribute name",
            ],
        ],
        "DbGetClassAttributeProperty": [
            [
                PyTango.DevVarStringArray,
                "Str[0] = Tango class name\nStr[1] = Attribute name\nStr[n] = Attribute name",
            ],
            [
                PyTango.DevVarStringArray,
                "Str[0] = Tango class name\nStr[1] = Attribute property  number\nStr[2] = Attribute property 1 name\nStr[3] = Attribute property 1 value\nStr[n + 1] = Attribute property 2 name\nStr[n + 2] = Attribute property 2 value",
            ],
        ],
        "DbGetClassAttributeProperty2": [
            [
                PyTango.DevVarStringArray,
                "Str[0] = Tango class name\nStr[1] = Attribute name\nStr[n] = Attribute name",
            ],
            [
                PyTango.DevVarStringArray,
                "Str[0] = Tango class name\nStr[1] = Attribute property  number\nStr[2] = Attribute property 1 name\nStr[3] = Attribute property 1 value number (array case)\nStr[4] = Attribute property 1 value\nStr[n] = Attribute property 1 value (array case)\nStr[n + 1] = Attribute property 2 name\nStr[n + 2] = Attribute property 2 value number (array case)\nStr[n + 3] = Attribute property 2 value\nStr[n + m] = Attribute property 2 value (array case)",
            ],
        ],
        "DbGetClassAttributePropertyHist": [
            [
                PyTango.DevVarStringArray,
                "Str[0] = Tango class\nStr[1] = Attribute name\nStr[2] = Property name",
            ],
            [
                PyTango.DevVarStringArray,
                "Str[0] = Attribute name\nStr[1] = Property name\nStr[2] = date\nStr[3] = Property value number (array case)\nStr[4] = Property value 1\nStr[n] = Property value n",
            ],
        ],
        "DbGetClassForDevice": [
            [PyTango.DevString, "Device name"],
            [PyTango.DevString, "Device Tango class"],
        ],
        "DbGetClassInheritanceForDevice": [
            [PyTango.DevString, "Device name"],
            [
                PyTango.DevVarStringArray,
                "Classes off the specified device.\n[0] - is the class of the device.\n[1] - is the class from the device class is inherited.\n........and so on",
            ],
        ],
        "DbGetClassList": [
            [PyTango.DevString, "Filter"],
            [PyTango.DevVarStringArray, "Class list"],
        ],
        "DbGetClassProperty": [
            [
                PyTango.DevVarStringArray,
                "Str[0] = Tango class\nStr[1] = Property name\nStr[2] = Property name",
            ],
            [
                PyTango.DevVarStringArray,
                "Str[0] = Tango class\nStr[1] = Property number\nStr[2] = Property name\nStr[3] = Property value number (array case)\nStr[4] = Property value\nStr[n] = Propery value (array case)\n....",
            ],
        ],
        "DbGetClassPropertyHist": [
            [PyTango.DevVarStringArray, "Str[0] = Tango class\nStr[1] = Property name"],
            [
                PyTango.DevVarStringArray,
                "Str[0] = Property name\nStr[1] = date\nStr[2] = Property value number (array case)\nStr[3] = Property value 1\nStr[n] = Property value n",
            ],
        ],
        "DbGetClassPropertyList": [
            [PyTango.DevString, "The filter"],
            [PyTango.DevVarStringArray, "Property name list"],
        ],
        "DbGetDeviceAlias": [
            [PyTango.DevString, "The device name"],
            [PyTango.DevString, "The alias found"],
        ],
        "DbGetDeviceAliasList": [
            [PyTango.DevString, "The filter"],
            [PyTango.DevVarStringArray, "Device alias list"],
        ],
        "DbGetDeviceAttributeList": [
            [PyTango.DevVarStringArray, "Str[0] = Device name\nStr[1] = Wildcard"],
            [PyTango.DevVarStringArray, "attribute name list"],
        ],
        "DbGetDeviceAttributeProperty": [
            [
                PyTango.DevVarStringArray,
                "Str[0] = Device name\nStr[1] = Attribute name\nStr[n] = Attribute name",
            ],
            [
                PyTango.DevVarStringArray,
                "Str[0] = Device name\nStr[1] = Attribute property  number\nStr[2] = Attribute property 1 name\nStr[3] = Attribute property 1 value\nStr[n + 1] = Attribute property 2 name\nStr[n + 2] = Attribute property 2 value",
            ],
        ],
        "DbGetDeviceAttributeProperty2": [
            [
                PyTango.DevVarStringArray,
                "Str[0] = Device name\nStr[1] = Attribute name\nStr[n] = Attribute name",
            ],
            [
                PyTango.DevVarStringArray,
                "Str[0] = Device name\nStr[1] = Attribute property  number\nStr[2] = Attribute property 1 name\nStr[3] = Attribute property 1 value number (array case)\nStr[4] = Attribute property 1 value\nStr[n] = Attribute property 1 value (array case)\nStr[n + 1] = Attribute property 2 name\nStr[n + 2] = Attribute property 2 value number (array case)\nStr[n + 3] = Attribute property 2 value\nStr[n + m] = Attribute property 2 value (array case)",
            ],
        ],
        "DbGetDeviceAttributePropertyHist": [
            [
                PyTango.DevVarStringArray,
                "Str[0] = Device name\nStr[1] = Attribute name\nStr[2] = Property name",
            ],
            [
                PyTango.DevVarStringArray,
                "Str[0] = Attribute name\nStr[1] = Property name\nStr[2] = date\nStr[3] = Property value number (array case)\nStr[4] = Property value 1\nStr[n] = Property value n",
            ],
        ],
        "DbGetDeviceClassList": [
            [PyTango.DevString, "Device server process name"],
            [
                PyTango.DevVarStringArray,
                "Str[0] = Device name\nStr[1] = Tango class\nStr[n] = Device name\nStr[n + 1] = Tango class",
            ],
        ],
        "DbGetDeviceDomainList": [
            [PyTango.DevString, "The wildcard"],
            [PyTango.DevVarStringArray, "Device name domain list"],
        ],
        "DbGetDeviceExportedList": [
            [PyTango.DevString, "filter"],
            [PyTango.DevVarStringArray, "list of exported devices"],
        ],
        "DbGetDeviceFamilyList": [
            [PyTango.DevString, "The wildcard"],
            [PyTango.DevVarStringArray, "Family list"],
        ],
        "DbGetDeviceInfo": [
            [PyTango.DevString, "Device name"],
            [
                PyTango.DevVarLongStringArray,
                "Str[0] = Device name\nStr[1] = CORBA IOR\nStr[2] = Device version\nStr[3] = Device Server name\nStr[4] = Device Server process host name\nStr[5] = Started date (or ? if not set)\nStr[6] = Stopped date (or ? if not set)\nStr[7] = Device class\n\nLg[0] = Device exported flag\nLg[1] = Device Server process PID (or -1 if not set)",
            ],
        ],
        "DbGetDeviceList": [
            [
                PyTango.DevVarStringArray,
                "argin[0] : server name\nargin[1] : class name",
            ],
            [
                PyTango.DevVarStringArray,
                "The list of devices for specified server and class.",
            ],
        ],
        "DbGetDeviceWideList": [
            [PyTango.DevString, "filter"],
            [PyTango.DevVarStringArray, "list of exported devices"],
        ],
        "DbGetDeviceMemberList": [
            [PyTango.DevString, "The filter"],
            [PyTango.DevVarStringArray, "Device names member list"],
        ],
        "DbGetDeviceProperty": [
            [
                PyTango.DevVarStringArray,
                "Str[0] = Device name\nStr[1] = Property name\nStr[n] = Property name",
            ],
            [
                PyTango.DevVarStringArray,
                "Str[0] = Device name\nStr[1] = Property number\nStr[2] = Property name\nStr[3] = Property value number (array case)\nStr[4] = Property value 1\nStr[n] = Property value n (array case)\nStr[n + 1] = Property name\nStr[n + 2] = Property value number (array case)\nStr[n + 3] = Property value 1\nStr[n + m] = Property value m",
            ],
        ],
        "DbGetDevicePropertyHist": [
            [PyTango.DevVarStringArray, "Str[0] = Device name\nStr[2] = Property name"],
            [
                PyTango.DevVarStringArray,
                "Str[0] = Property name\nStr[1] = date\nStr[2] = Property value number (array case)\nStr[3] = Property value 1\nStr[n] = Property value n",
            ],
        ],
        "DbGetDevicePropertyList": [
            [PyTango.DevVarStringArray, "Str[0] = device name\nStr[1] = Filter"],
            [PyTango.DevVarStringArray, "Property name list"],
        ],
        "DbGetDeviceServerClassList": [
            [PyTango.DevString, "device server process name"],
            [PyTango.DevVarStringArray, "list of classes for this device server"],
        ],
        "DbGetExportdDeviceListForClass": [
            [PyTango.DevString, "Class name"],
            [PyTango.DevVarStringArray, "Device exported list"],
        ],
        "DbGetHostList": [
            [PyTango.DevString, "The filter"],
            [PyTango.DevVarStringArray, "Host name list"],
        ],
        "DbGetHostServerList": [
            [PyTango.DevString, "The filter"],
            [PyTango.DevVarStringArray, "Device server process name list"],
        ],
        "DbGetHostServersInfo": [
            [PyTango.DevString, "Host name"],
            [
                PyTango.DevVarStringArray,
                "Server info for all servers running on specified host",
            ],
        ],
        "DbGetInstanceNameList": [
            [PyTango.DevString, "Server name"],
            [
                PyTango.DevVarStringArray,
                "The instance names found for specified server.",
            ],
        ],
        "DbGetObjectList": [
            [PyTango.DevString, "The filter"],
            [PyTango.DevVarStringArray, "Object name list"],
        ],
        "DbGetProperty": [
            [
                PyTango.DevVarStringArray,
                "Str[0] = Object name\nStr[1] = Property name\nStr[n] = Property name",
            ],
            [
                PyTango.DevVarStringArray,
                "Str[0] = Object name\nStr[1] = Property number\nStr[2] = Property name\nStr[3] = Property value number (array case)\nStr[4] = Property value 1\nStr[n] = Property value n (array case)\nStr[n + 1] = Property name\nStr[n + 2] = Property value number (array case)\nStr[n + 3] = Property value 1\nStr[n + m] = Property value m",
            ],
        ],
        "DbGetPropertyHist": [
            [PyTango.DevVarStringArray, "Str[0] = Object name\nStr[2] = Property name"],
            [
                PyTango.DevVarStringArray,
                "Str[0] = Property name\nStr[1] = date\nStr[2] = Property value number (array case)\nStr[3] = Property value 1\nStr[n] = Property value n",
            ],
        ],
        "DbGetPropertyList": [
            [PyTango.DevVarStringArray, "Str[0] = Object name\nStr[1] = filter"],
            [PyTango.DevVarStringArray, "Property name list"],
        ],
        "DbGetServerInfo": [
            [PyTango.DevString, "server name"],
            [PyTango.DevVarStringArray, "server info"],
        ],
        "DbGetServerList": [
            [PyTango.DevString, "The filter"],
            [PyTango.DevVarStringArray, "Device server process name list"],
        ],
        "DbGetServerNameList": [
            [PyTango.DevString, "wildcard for server names."],
            [PyTango.DevVarStringArray, "server names found."],
        ],
        "DbImportDevice": [
            [PyTango.DevString, "Device name (or alias)"],
            [
                PyTango.DevVarLongStringArray,
                "Str[0] = device name\nStr[1] = CORBA IOR\nStr[2] = device version\nStr[3] = device server process name\nStr[4] = host name\nStr[5] = Tango class name\n\nLg[0] = Exported flag\nLg[1] = Device server process PID",
            ],
        ],
        "DbImportEvent": [
            [PyTango.DevString, "name of event channel or factory"],
            [PyTango.DevVarLongStringArray, "export information e.g. IOR"],
        ],
        "DbInfo": [
            [PyTango.DevVoid, "none"],
            [
                PyTango.DevVarStringArray,
                "Miscellaneous info like:\n- Device defined in database\n- Device marked as exported in database\n- Device server process defined in database\n- Device server process marked as exported in database\n- Device properties defined in database\n- Class properties defined in database\n- Device attribute properties defined in database\n- Class attribute properties defined in database\n- Object properties defined in database",
            ],
        ],
        "DbPutAttributeAlias": [
            [
                PyTango.DevVarStringArray,
                "Str[0] = attribute name\nStr[1] = attribute alias",
            ],
            [PyTango.DevVoid, "none"],
        ],
        "DbPutClassAttributeProperty": [
            [
                PyTango.DevVarStringArray,
                "Str[0] = Tango class name\nStr[1] = Attribute number\nStr[2] = Attribute name\nStr[3] = Property number\nStr[4] = Property name\nStr[5] = Property value\n.....",
            ],
            [PyTango.DevVoid, "none"],
        ],
        "DbPutClassAttributeProperty2": [
            [
                PyTango.DevVarStringArray,
                "Str[0] = Tango class name\nStr[1] = Attribute number\nStr[2] = Attribute name\nStr[3] = Property number\nStr[4] = Property name\nStr[5] = Property value number (array case)\nStr[5] = Property value 1\nStr[n] = Property value n (array case)\n.....",
            ],
            [PyTango.DevVoid, "none"],
        ],
        "DbPutClassProperty": [
            [
                PyTango.DevVarStringArray,
                "Str[0] = Tango class name\nStr[1] = Property number\nStr[2] = Property name\nStr[3] = Property value number\nStr[4] = Property value 1\nStr[n] = Property value n\n....",
            ],
            [PyTango.DevVoid, "none"],
        ],
        "DbPutDeviceAlias": [
            [PyTango.DevVarStringArray, "Str[0] = device name\nStr[1] = alias name"],
            [PyTango.DevVoid, "none"],
        ],
        "DbPutDeviceAttributeProperty": [
            [
                PyTango.DevVarStringArray,
                "Str[0] = Device name\nStr[1] = Attribute number\nStr[2] = Attribute name\nStr[3] = Property number\nStr[4] = Property name\nStr[5] = Property value\n.....",
            ],
            [PyTango.DevVoid, "none"],
        ],
        "DbPutDeviceAttributeProperty2": [
            [
                PyTango.DevVarStringArray,
                "Str[0] = Device name\nStr[1] = Attribute number\nStr[2] = Attribute name\nStr[3] = Property number\nStr[4] = Property name\nStr[5] = Property value number (array case)\nStr[5] = Property value 1\nStr[n] = Property value n (array case)\n.....",
            ],
            [PyTango.DevVoid, "none"],
        ],
        "DbPutDeviceProperty": [
            [
                PyTango.DevVarStringArray,
                "Str[0] = Tango device name\nStr[1] = Property number\nStr[2] = Property name\nStr[3] = Property value number\nStr[4] = Property value 1\nStr[n] = Property value n\n....",
            ],
            [PyTango.DevVoid, "none"],
        ],
        "DbPutProperty": [
            [
                PyTango.DevVarStringArray,
                "Str[0] = Object name\nStr[1] = Property number\nStr[2] = Property name\nStr[3] = Property value number\nStr[4] = Property value 1\nStr[n] = Property value n\n....",
            ],
            [PyTango.DevVoid, "none"],
        ],
        "DbPutServerInfo": [
            [PyTango.DevVarStringArray, "server info"],
            [PyTango.DevVoid, "none"],
        ],
        "DbUnExportDevice": [
            [PyTango.DevString, "Device name"],
            [PyTango.DevVoid, "none"],
        ],
        "DbUnExportEvent": [
            [PyTango.DevString, "name of event channel or factory to unexport"],
            [PyTango.DevVoid, "none"],
        ],
        "DbUnExportServer": [
            [PyTango.DevString, "Device server name (executable/instance)"],
            [PyTango.DevVoid, "none"],
        ],
        "ResetTimingValues": [[PyTango.DevVoid, "none"], [PyTango.DevVoid, "none"]],
        "DbGetDataForServerCache": [
            [
                PyTango.DevVarStringArray,
                "Elt[0] = DS name (exec_name/inst_name), Elt[1] = Host name",
            ],
            [
                PyTango.DevVarStringArray,
                "All the data needed by the device server during its startup sequence. Precise list depend on the device server",
            ],
        ],
        "DbDeleteAllDeviceAttributeProperty": [
            [
                PyTango.DevVarStringArray,
                "str[0] = device name\nStr[1]...str[n] = attribute name(s)",
            ],
            [PyTango.DevVoid, "none"],
        ],
        "DbMySqlSelect": [
            [PyTango.DevString, "MySql Select command"],
            [
                PyTango.DevVarLongStringArray,
                "MySql Select command result\n - svalues : select results\n - lvalue[n] : =0 if svalue[n] is null else =1\n (last lvalue -1) is number of rows, (last lvalue) is number of fields",
            ],
        ],
        "DbGetCSDbServerList": [
            [PyTango.DevVoid, "none"],
            [
                PyTango.DevVarStringArray,
                "List of host:port with one element for each database server",
            ],
        ],
        "DbGetAttributeAlias2": [
            [PyTango.DevString, "The attribute name (dev_name/att_name)"],
            [PyTango.DevString, "The attribute alias name (or empty string)"],
        ],
        "DbGetAliasAttribute": [
            [PyTango.DevString, "The attribute alias"],
            [PyTango.DevString, "The attribute name (dev_name/att_name)"],
        ],
    }

    #    Attribute definitions
    attr_list = {
        "StoredProcedureRelease": [[PyTango.DevString, PyTango.SCALAR, PyTango.READ]],
        "Timing_average": [[PyTango.DevDouble, PyTango.SPECTRUM, PyTango.READ, 64]],
        "Timing_minimum": [[PyTango.DevDouble, PyTango.SPECTRUM, PyTango.READ, 64]],
        "Timing_maximum": [[PyTango.DevDouble, PyTango.SPECTRUM, PyTango.READ, 64]],
        "Timing_calls": [[PyTango.DevDouble, PyTango.SPECTRUM, PyTango.READ, 64]],
        "Timing_index": [[PyTango.DevString, PyTango.SPECTRUM, PyTango.READ, 64]],
        "Timing_info": [[PyTango.DevString, PyTango.SPECTRUM, PyTango.READ, 64]],
    }

    # ------------------------------------------------------------------
    #    DataBaseClass Constructor
    # ------------------------------------------------------------------
    def __init__(self, name):
        PyTango.DeviceClass.__init__(self, name)
        self.set_type(name)
        print("In DataBase Class  constructor")


# ==================================================================
#
#    DataBase class main method
#
# ==================================================================
def main():
    try:
        py = PyTango.Util(sys.argv)
        py.add_class(DataBaseClass, DataBase, "DataBase")

        U = PyTango.Util.instance()
        U.server_init()
        U.server_run()

    except PyTango.DevFailed as e:
        print("-------> Received a DevFailed exception:", e)
    except Exception as e:
        print("-------> An unforeseen exception occured....", e)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""
This is an internal PyTango module.
"""

__all__ = ("NumpyType", "numpy_type", "numpy_spectrum", "numpy_image")

__docformat__ = "restructuredtext"

import collections.abc
import numpy

from tango import Except, Attribute, AttributeInfo
from tango import CmdArgType as ArgType

from tango.attribute_proxy import AttributeProxy


def _numpy_invalid(*args, **kwds):
    Except.throw_exception(
        "PyTango_InvalidConversion",
        "There's no registered conversor to numpy.",
        "NumpyType.tango_to_numpy",
    )


class NumpyType:
    DevShort = numpy.int16
    DevLong = numpy.int32
    DevDouble = numpy.float64
    DevFloat = numpy.float32
    DevBoolean = numpy.bool_
    DevUShort = numpy.uint16
    DevULong = numpy.uint32
    DevUChar = numpy.ubyte
    DevLong64 = numpy.int64
    DevULong64 = numpy.uint64

    mapping = {
        ArgType.DevShort: DevShort,
        ArgType.DevLong: DevLong,
        ArgType.DevDouble: DevDouble,
        ArgType.DevFloat: DevFloat,
        ArgType.DevBoolean: DevBoolean,
        ArgType.DevUShort: DevUShort,
        ArgType.DevULong: DevULong,
        ArgType.DevUChar: DevUChar,
        ArgType.DevLong64: DevLong64,
        ArgType.DevULong: DevULong64,
    }

    @staticmethod
    def tango_to_numpy(param):
        if isinstance(param, ArgType):
            tg_type = param
        if isinstance(param, AttributeInfo):  # or AttributeInfoEx
            tg_type = param.data_type
        elif isinstance(param, Attribute):
            tg_type = param.get_data_type()
        elif isinstance(param, AttributeProxy):
            tg_type = param.get_config().data_type
        else:
            tg_type = param
        try:
            return NumpyType.mapping[tg_type]
        except Exception:
            _numpy_invalid()

    @staticmethod
    def spectrum(tg_type, dim_x):
        """
        numpy_spectrum(self, tg_type, dim_x, dim_y) -> numpy.array
        numpy_spectrum(self, tg_type, sequence) -> numpy.array

                Get a square numpy array to be used with tango.
                One version gets dim_x and creates an object with
                this size. The other version expects any sequence to
                convert.

            Parameters:
                - tg_type : (ArgType): The tango type. For convenience, it
                            can also extract this information from an
                            Attribute, AttributeInfo or AttributeProxy
                            object.
                - dim_x : (int)
                - sequence:
        """
        np_type = NumpyType.tango_to_numpy(tg_type)
        if isinstance(dim_x, collections.abc.Sequence):
            return numpy.array(dim_x, dtype=np_type)
        else:
            return numpy.ndarray(shape=(dim_x,), dtype=np_type)

    @staticmethod
    def image(tg_type, dim_x, dim_y=None):
        """
        numpy_image(self, tg_type, dim_x, dim_y) -> numpy.array
        numpy_image(self, tg_type, sequence) -> numpy.array

                Get a square numpy array to be used with tango.
                One version gets dim_x and dim_y and creates an object with
                this size. The other version expects a square sequence of
                sequences to convert.

            Parameters:
                - tg_type : (ArgType): The tango type. For convenience, it
                            can also extract this information from an
                            Attribute, AttributeInfo or AttributeProxy
                            object.
                - dim_x : (int)
                - dim_y : (int)
                - sequence:
        """
        np_type = NumpyType.tango_to_numpy(tg_type)
        if dim_y is None:
            return numpy.array(dim_x, dtype=np_type)
        else:
            return numpy.ndarray(
                shape=(
                    dim_y,
                    dim_x,
                ),
                dtype=np_type,
            )


numpy_spectrum = NumpyType.spectrum
numpy_image = NumpyType.image
numpy_type = NumpyType.tango_to_numpy



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later
from __future__ import annotations

import time

try:
    import numpy.typing as npt
except ImportError:
    npt = None

from tango import (
    AttrDataFormat,
    AttrQuality,
    AttrWriteType,
    DevState,
)
from tango.server import Device
from tango.server import command, attribute
from tango.test_utils import (
    DeviceTestContext,
    assert_close,
    check_attr_type,
)


def test_scalar_attribute_declared_with_typing():

    class TestDevice(Device):
        attr_value = None

        hint_scalar: int = attribute(
            access=AttrWriteType.READ_WRITE, fget="read_attr", fset="write_attr"
        )

        hint_scalar_quality: tuple[int, float, AttrQuality] = attribute(
            access=AttrWriteType.READ_WRITE, fget="read_attr", fset="write_attr"
        )

        hint_with_tuple_spectrum: tuple[int, int, int] = attribute(
            access=AttrWriteType.READ_WRITE, fget="read_attr", fset="write_attr"
        )

        hint_with_tuple_quality_spectrum: tuple[
            tuple[int, int, int], float, AttrQuality
        ] = attribute(
            access=AttrWriteType.READ_WRITE,
            fget="read_attr_with_quality",
            fset="write_attr",
        )

        hint_with_list_spectrum: list[int] = attribute(
            access=AttrWriteType.READ_WRITE,
            max_dim_x=3,
            max_dim_y=3,
            fget="read_attr",
            fset="write_attr",
        )

        hint_with_tuple_image: tuple[
            tuple[int, int, int], tuple[int, int, int], tuple[int, int, int]
        ] = attribute(
            access=AttrWriteType.READ_WRITE, fget="read_attr", fset="write_attr"
        )

        hint_with_tuple_quality_image: tuple[
            tuple[tuple[int, int, int], tuple[int, int, int], tuple[int, int, int]],
            float,
            AttrQuality,
        ] = attribute(
            access=AttrWriteType.READ_WRITE,
            fget="read_attr_with_quality",
            fset="write_attr",
        )

        def read_attr(self):
            return self.attr_value

        def read_attr_with_quality(self):
            return self.attr_value, time.time(), AttrQuality.ATTR_VALID

        def write_attr(self, value):
            self.attr_value = value

        @attribute(access=AttrWriteType.READ_WRITE)
        def attribute_hint(self) -> int:
            return self.attr_value

        @attribute_hint.write
        def attribute_hint(self, value):
            self.attr_value = value

        @attribute(access=AttrWriteType.READ_WRITE)
        def attribute_hint_quality(self) -> tuple[int, float, AttrQuality]:
            return self.attr_value

        @attribute_hint_quality.write
        def attribute_hint_quality(self, value):
            self.attr_value = value

        @command()
        def reset(self):
            self.attr_value = None

    def check_attribute_with_size(proxy, attr, value):
        setattr(proxy, attr, value)
        assert_close(getattr(proxy, attr), value)
        proxy.reset()

    with DeviceTestContext(TestDevice) as proxy:
        check_attribute_with_size(proxy, "hint_scalar", 1)
        check_attribute_with_size(proxy, "hint_scalar_quality", 2)
        check_attribute_with_size(proxy, "hint_with_tuple_spectrum", [1, 2, 3])
        check_attribute_with_size(proxy, "hint_with_tuple_quality_spectrum", [1, 2, 3])
        check_attribute_with_size(proxy, "hint_with_list_spectrum", [1, 2, 3])
        check_attribute_with_size(
            proxy, "hint_with_tuple_image", [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        )
        check_attribute_with_size(
            proxy, "hint_with_tuple_quality_image", [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        )
        check_attribute_with_size(proxy, "attribute_hint", 1)
        check_attribute_with_size(proxy, "attribute_hint_quality", 2)


def test_attribute_self_typed_with_not_defined_name():
    _value = [None]

    def non_bound_read(device: "TestDevice") -> int:
        return _value[0]

    def non_bound_read_no_return_hint(device: "TestDevice"):
        return _value[0]

    def non_bound_write(device: "TestDevice", val_in: int):
        _value[0] = val_in

    class TestDevice(Device):
        _value = None

        assignment_attr: int = attribute(
            access=AttrWriteType.READ_WRITE, fget="read_attr", fset="write_attr"
        )

        non_bound_attr: int = attribute(
            access=AttrWriteType.READ_WRITE, fget=non_bound_read, fset=non_bound_write
        )

        non_bound_attr_in_write: int = attribute(
            access=AttrWriteType.READ_WRITE,
            fget=non_bound_read_no_return_hint,
            fset=non_bound_write,
        )

        def read_attr(self: "TestDevice"):
            return self._value

        def write_attr(self: "TestDevice", val_in):
            self._value = val_in

        @attribute
        def decorator_attr(self: "TestDevice") -> int:
            return self._value

        @decorator_attr.write
        def set_value(self: "TestDevice", val_in: int):
            self._value = val_in

        @attribute
        def decorator_attr_def_in_write(self: "TestDevice"):
            return self._value

        @decorator_attr_def_in_write.write
        def set_value_2(self: "TestDevice", val_in: int):
            self._value = val_in

    with DeviceTestContext(TestDevice) as proxy:
        proxy.assignment_attr = 1
        assert 1 == proxy.assignment_attr
        proxy.decorator_attr = 2
        assert 2 == proxy.decorator_attr
        proxy.decorator_attr_def_in_write = 3
        assert 3 == proxy.decorator_attr_def_in_write

        proxy.non_bound_attr = 1
        assert 1 == proxy.non_bound_attr
        proxy.non_bound_attr_in_write = 2
        assert 2 == proxy.non_bound_attr_in_write


if npt:

    def test_attribute_declared_with_numpy_typing():

        class TestDevice(Device):
            attr_value = None

            statement_declaration: npt.NDArray[int] = attribute(
                access=AttrWriteType.READ_WRITE,
                fget="read_attr",
                fset="write_attr",
                dformat=AttrDataFormat.SPECTRUM,
                max_dim_x=2,
                max_dim_y=2,
            )

            def read_attr(self):
                return self.attr_value

            def write_attr(self, value):
                self.attr_value = value

            def initialize_dynamic_attributes(self):
                attr = attribute(
                    name="dynamic_declaration",
                    access=AttrWriteType.READ_WRITE,
                    fget=self.read_dynamic_attr,
                    fset=self.write_dynamic_attr,
                    dformat=AttrDataFormat.SPECTRUM,
                    max_dim_x=2,
                    max_dim_y=2,
                )
                self.add_attribute(attr)

            def read_dynamic_attr(self, attr):
                return self.attr_value

            def write_dynamic_attr(self, attr):
                self.attr_value = attr.get_write_value()

            @attribute(
                access=AttrWriteType.READ_WRITE,
                dformat=AttrDataFormat.SPECTRUM,
                max_dim_x=2,
                max_dim_y=2,
            )
            def decorator_declaration(self) -> npt.NDArray[int]:
                return self.attr_value

            @decorator_declaration.write
            def decorator_declaration_write(self, value: npt.NDArray[int]):
                self.attr_value = value

            @command()
            def reset(self):
                self.attr_value = None

        def check_attribute(proxy, attr, value):
            setattr(proxy, attr, value)
            assert_close(getattr(proxy, attr), value)
            proxy.reset()

        with DeviceTestContext(TestDevice) as proxy:
            check_attribute(proxy, "statement_declaration", [1, 2])
            check_attribute(proxy, "decorator_declaration", [3, 4])
            check_attribute(proxy, "dynamic_declaration", [5, 6])


def test_devstate_attribute_declared_with_typing():

    class TestDevice(Device):

        @attribute()
        def state_attr(self) -> DevState:
            return DevState.MOVING

    with DeviceTestContext(TestDevice) as proxy:
        state_value = proxy.state_attr
        assert state_value == DevState.MOVING
        check_attr_type(state_value, AttrDataFormat.SCALAR, DevState)



# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

import pytest
from tango import AttributeConfig, CmdArgType, DevError, DevFailed, Group, ClientAddr
from tango._tango import GroupAttrReply, GroupCmdReply, GroupReply
from tango.pytango_pprint import _SEQUENCE_TYPES, _STRUCT_TYPES
from tango.server import Device, attribute, command
from tango.test_context import DeviceTestContext


STRUCTS_NOT_INSTANTIABLE_FROM_PYTHON = (
    GroupAttrReply,
    GroupCmdReply,
    GroupReply,
)
TESTABLE_TYPES = set(_STRUCT_TYPES).difference(STRUCTS_NOT_INSTANTIABLE_FROM_PYTHON)


class GroupTestDevice(Device):
    @attribute
    def attr(self) -> float:
        return 12.3

    @attr.setter
    def attr(self, val: float) -> None:
        pass

    @attribute
    def attr_fail(self) -> float:
        raise RuntimeError("Fail for test")

    @attr_fail.setter
    def attr_fail(self, val: float) -> None:
        raise RuntimeError("Fail for test")

    @command
    def cmd_fail(self):
        raise RuntimeError("Fail for test")


@pytest.fixture
def group():
    with DeviceTestContext(GroupTestDevice, device_name="test/device/1"):
        group_client = Group("test")
        group_client.add("test/device/1")
        yield group_client


@pytest.mark.parametrize(
    "to_string, requires_newline",
    [
        (str, True),
        (repr, False),
    ],
)
def test_pprint_structs(to_string, requires_newline):
    for struct_type in TESTABLE_TYPES:
        struct = struct_type()
        s = to_string(struct)
        assert struct_type.__name__ in s
        assert ("\n" in s) == requires_newline


@pytest.mark.parametrize(
    "to_string, requires_newline",
    [
        (str, True),
        (repr, False),
    ],
)
def test_pprint_group_structs(to_string, requires_newline, group):
    group_cmd_reply = group.command_inout("State")[0]
    group_attr_reply = group.read_attribute("attr")[0]
    group_reply = group.write_attribute("attr", 45.6)[0]
    group_cmd_failed_reply = group.command_inout("cmd_fail")[0]
    group_attr_failed_reply = group.read_attribute("attr_fail")[0]
    group_failed_reply = group.write_attribute("attr_fail", 45.6)[0]
    for struct in [
        group_cmd_reply,
        group_attr_reply,
        group_reply,
        group_cmd_failed_reply,
        group_attr_failed_reply,
        group_failed_reply,
    ]:
        s = to_string(struct)
        assert type(struct).__name__ in s
        assert ("\n" in s) == requires_newline


@pytest.mark.parametrize(
    "to_string, requires_newline",
    [
        (str, True),
        (repr, False),
    ],
)
def test_pprint_dev_failed_dev_error(to_string, requires_newline):
    dev_failed = DevFailed(DevError(), DevError())
    s = to_string(dev_failed)
    assert "DevFailed" in s
    assert "DevError" in s
    assert ("\n" in s) == requires_newline


@pytest.mark.parametrize("to_string", [str, repr])
def test_pprint_sequences(to_string):
    for seq_type in _SEQUENCE_TYPES:
        struct = seq_type()
        s = to_string(struct)
        assert "[" in s
        assert "]" in s


def test_pprint_data_type_is_cmd_arg_type():
    config = AttributeConfig()
    for dtype_value in CmdArgType.values:
        config.data_type = dtype_value
        s = str(config)
        assert f"data_type = {CmdArgType(dtype_value)!r}" in s


def test_pprint_client_addr():
    addr = ClientAddr("127.0.0.1")
    s = str(addr)
    assert s == "Client identification not available"
    r = repr(addr)
    assert "ClientAddr" in r
    assert "127.0.0.1" in r



#!/usr/bin/env python3
# SPDX-FileCopyrightText: All Contributors to the PyTango project
# SPDX-License-Identifier: LGPL-3.0-or-later

"""Trivial example of a high-level API device server."""

from tango.server import Device, attribute


class Motor(Device):
    def init_device(self):
        super().init_device()
        self.attr_position_read = 1.0

    @attribute(dtype=float)
    def position(self):
        return self.attr_position_read


def main():
    Motor.run_server()


if __name__ == "__main__":
    main()
